/*

Copyright 2001 Ionific Ltd.

This file is part of PlusV Reference Software.

PlusV Reference Software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

PlusV Reference Software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PlusV Reference Software; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
/**
 * @file        ilmp3.h
 * @guid        d0ddd7a4b327419d95840d52a4f10aa7
 * @copyright   2001 Ionific Ltd.
 * @author      Jussi Lahdenniemi, jussi.lahdenniemi@ionific.com
 * @project     plusv/interleave
 * @description MP3 reader of the +V interleaver
 * @note        This file is not intended as an example or anything. It is
 *              a rip-off of the Ionific's MP3 decoder; only the bitstream
 *              decoding remains. Interleaving +V data into an existing
 *              MP3 bitstream requires the MP3 bistream to be decoded before
 *              the correct insertion point is known.
 */

#ifndef Hc24fcb9913ed4d6a875e606f3f1f1dc3
#define Hc24fcb9913ed4d6a875e606f3f1f1dc3

#include <stddef.h>

#define MP3_OK          0
#define MP3_END         1
#define MP3_ERROR       -1
#define MP3_INVALID     -2

typedef void* mp3Stream_t;

typedef struct
{
    unsigned    mpegVersion;    /* 10 for 1.0, 20 for 2.0 or 25 for 2.5 */
    unsigned    layer;          /* Currently always 3 */
    unsigned    bitrate;        /* The stream's bit rate (kbps) */
    unsigned    frequency;      /* The stream's frequency (Hz) */
    unsigned    channels;       /* Number of channels in stream */
    int         plusV;          /* Stream's PlusV version, or -1 if no plusV */

} mp3StreamInfo_t;

typedef size_t (*streamRequest_t)( void* info, void* buffer, size_t size );

void mp3InitDecoder( void* (*allocator)( size_t ),
                     void (*freer)( void* ));
mp3Stream_t mp3OpenStream( streamRequest_t request, void* reqInfo,
                           void* x, void* y );
void mp3CloseStream( mp3Stream_t stream );
int mp3GetStreamInfo( mp3Stream_t stream, mp3StreamInfo_t* info );
int mp3DecodeFrame( mp3Stream_t stream );

typedef struct
{
    streamRequest_t     requester;
    void*               reqInfo;

    unsigned char*      buffer;
    int                 bufAllocated;
    size_t              maxlen;
    size_t              avail;
    size_t              pos;

    unsigned long       ibuf[2];
    size_t              ibits;

    size_t              reserve;

    unsigned char*      copyBuf;
    size_t              copyLen;
    size_t              copyPos;
    unsigned            copyBit;
}                       bitsStream_t;

void bitsInit( bitsStream_t* bits, size_t maxlength, unsigned char *buffer,
               streamRequest_t req, void* reqInfo );
void bitsFree( bitsStream_t* bits );
unsigned long bitsRead( bitsStream_t* bits, size_t bitcount, int* failed );
unsigned long bitsPeek( bitsStream_t* bits, size_t bitcount, int* failed );
unsigned long bitsRead1( bitsStream_t* bits, int *failed );
void bitsDiscard( bitsStream_t* bits, size_t count, int *failed );
int bitsSetPosition( bitsStream_t* bits, size_t bytes );
int bitsRewind( bitsStream_t* bits, size_t count );
size_t bitsAvail( bitsStream_t* bits );
void bitsReserve( bitsStream_t* bits, size_t count );
int bitsUpdate( bitsStream_t* bits );
void bitsAlignOctet( bitsStream_t* bits );
void bitsStartCopy( bitsStream_t* bits, void* buffer, size_t length );
size_t bitsEndCopy( bitsStream_t* bits );

typedef struct
{
    unsigned islut;
    unsigned linbits;
    unsigned xlen, ylen;
    unsigned bits;
    union
    {
        unsigned const *table;
        unsigned const *tree;
    }       u;
} mp3HuffmanTable_t;

extern mp3HuffmanTable_t const huffmanTable_g[];
extern mp3HuffmanTable_t const huffmanTable_quadA_g;

enum
{
    STREAM_INITIALIZED,
    STREAM_HEADER_READ,
    STREAM_DECODING,
    STREAM_INVALID,
};

#define MP3_FRAME_LENGTH    1152
#define MP3_HEADER_BITS     32
#define BYTE_ORDER_LITTLE_ENDIAN

#define QUANTIZSIZE         256
#define PRECISE_ROUND

#define MP3_HDR_SYNC( hdr )        ( ((unsigned)(hdr)[0] << 3 ) | (unsigned)(((hdr)[1] >> 5 ) & 0x7 ))
#define MP3_HDR_MPEG10( hdr )      ( ((hdr)[1] >>  4 ) & 0x1 )
#define MP3_HDR_ID( hdr )          ( ((hdr)[1] >>  3 ) & 0x1 )
#define MP3_HDR_LAYER( hdr )       ( ((hdr)[1] >>  1 ) & 0x3 )
#define MP3_HDR_PROTECTION( hdr )  (  (hdr)[1]         & 0x1 )
#define MP3_HDR_BITRATE_IDX( hdr ) ( ((hdr)[2] >>  4 ) & 0xf )
#define MP3_HDR_FREQUENCY( hdr )   ( ((hdr)[2] >>  2 ) & 0x3 )
#define MP3_HDR_PADDING( hdr )     ( ((hdr)[2] >>  1 ) & 0x1 )
#define MP3_HDR_PRIV( hdr )        (  (hdr)[2]         & 0x1 )
#define MP3_HDR_MODE( hdr )        ( ((hdr)[3] >>  6 ) & 0x3 )
#define MP3_HDR_MODE_EXT( hdr )    ( ((hdr)[3] >>  4 ) & 0x3 )
#define MP3_HDR_COPYRIGHT( hdr )   ( ((hdr)[3] >>  3 ) & 0x1 )
#define MP3_HDR_ORIGINAL( hdr )    ( ((hdr)[3] >>  2 ) & 0x1 )
#define MP3_HDR_EMPHASIS( hdr )    (  (hdr)[3]         & 0x3 )

#define MP3_MODE_STEREO     0
#define MP3_MODE_JOINT      1
#define MP3_MODE_DUAL       2
#define MP3_MODE_SINGLE     3

typedef struct
{
    unsigned                magic;
    streamRequest_t         requester;
    void*                   reqInfo;
    int                     status;
    unsigned char           header[4];
    mp3StreamInfo_t         info;
    int                     plusvSeen;
    bitsStream_t            bits;
    struct
    {
        unsigned            main_data_begin;
        unsigned            scfsi[2][4];
        unsigned            part2_3_length[2][2];
        unsigned            big_values[2][2];
        unsigned            global_gain[2][2];
        unsigned            scalefac_compress[2][2];
        unsigned            window_switching_flag[2][2];
        unsigned            block_type[2][2];
        unsigned            mixed_block_flag[2][2];
        unsigned            table_select[2][2][3];
        unsigned            subblock_gain[2][2][3];
        unsigned            region_count[2][2][2];
        unsigned            preflag[2][2];
        unsigned            scalefac_scale[2][2];
        unsigned            count1table_select[2][2];

    }                       l3side;
    unsigned char           _sideinfo[32];
    unsigned                _sidelength;
    struct
    {
        unsigned            l[2][23];
        unsigned            s[2][2][13][3];

    }                       scalefac;
    struct
    {
        unsigned            l;
        unsigned            s[3];

    }                       lastNonzeroBand;
} mp3State_t;
#define MP3_MAGIC 0xdec0de13
extern unsigned const bitrates[2][4][16];
void mp3InitState( mp3State_t* state );
void mp3FreeState( mp3State_t* state );
size_t requestData( mp3State_t* str, void* buffer, size_t minsz, size_t maxsz );
int mp3DecodeLayerIII( mp3State_t* state );

#endif // Hc24fcb9913ed4d6a875e606f3f1f1dc3

