/*

Copyright 2001 Ionific Ltd.

This file is part of PlusV Reference Software.

PlusV Reference Software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

PlusV Reference Software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PlusV Reference Software; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
/**
 * @file        interleave.c
 * @guid        641acdd17b7648449789598a1bcc9cfb
 * @copyright   2001 Ionific Ltd.
 * @author      Jussi Lahdenniemi, jussi.lahdenniemi@ionific.com
 * @project     plusv/interleave
 * @description Interleave main file
 *              Interleaves MP3 and PlusV bitstreams into a new MP3 bitstream
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "ilmp3.h"

/* Bitrate use table, for statistics after encoding */

static unsigned bruse[16];

/**
 * @function    readpv
 * @guid        35d7d2d252fb44f4b74485122eac30d5
 * @description Reads a frame of PlusV information from bitstream
 * @param       in {FILE*, in} Input file
 * @param       buffer {unsigned char*, in} Buffer for the frame
 * @param       pvpermp3 {unsigned*, out} PlusV frames per MP3 frame
 * @return      {unsigned} Length of the frame read
 * @value       0 if no frame available
 */
unsigned readpv( FILE* in, unsigned char* buffer, unsigned* pvpermp3 )
{
    static int counter = -1;

    unsigned char c;
    if( fread( &c, 1, 1, in ) != 1 )
    {
        /* End of file */

        return 0;
    }

    if( c == 0xd6 )
    {
        /* FDB? */

        unsigned char c2[4];
        if( fread( c2, 1, 4, in ) != 4 )
        {
            fprintf( stderr, "Error in the PlusV bitstream\n" );
            return 0;
        }

        if( c2[0] == 0x53 && c2[1] == 0xd6 && c2[2] == 0xfd &&
            ( c2[3] & 0xc0 ) == 0 && ( c2[3] & 0x3f ) >= 3 )
        {
            unsigned len, rlen;

            /* FDB. */

            buffer[0] = c;
            memcpy( buffer + 1, c2, 4 );
            len = ( c2[3] & 0x3f ) - 5;
            rlen = fread( buffer + 5, 1, len, in );

            /* Check fPrivate bit 0 */

            if( buffer[6] & 0x40 )
            {
                *pvpermp3 = 3;
            }
            else
            {
                *pvpermp3 = 2;
            }

            if( rlen < len )
            {
                fprintf( stderr, "Error in the PlusV bitstream\n" );
                return 0;
            }

            counter = 1;
            return rlen + 5;
        }
        else
        {
            fprintf( stderr, "Error in the PlusV bitstream\n" );
            return 0;
        }
    }
    else
    if((int)( c >> 6 ) == counter &&
       ( c & 0x3f ) >= 3 )
    {
        /* LDB */

        unsigned len = ( c & 0x3f ) - 1;
        unsigned rlen;

        buffer[0] = c;
        rlen = fread( buffer + 1, 1, len, in );

        if( rlen < len )
        {
            fprintf( stderr, "Error in the PlusV bitstream\n" );
            return 0;
        }

        counter++;
        if( counter == 3 ) counter = 0;
        return rlen + 1;
    }
    else
    {
        fprintf( stderr, "Error in the PlusV bitstream (%02x, %u)\n",
                 c, counter );
        return 0;
    }
}

/**
 * @function    request
 * @guid        ea9440a2467c4842ab361eb16a9fbb97
 * @description Data request callback for the MP3 decoder
 * @param       in {FILE*, in} The input file
 * @param       buffer {void*, in} Buffer for the data
 * @param       size {size_t, in} Size of the buffer
 * @return      {size_t} Number of bytes read from the file
 */
size_t request( FILE* in, void* buffer, size_t size )
{
    if( buffer == NULL )
    {
        fseek( in, size, SEEK_CUR );
        return size;
    }
    else
    {
        return fread( buffer, 1, size, in );
    }
}

/**
 * @function    writemp3pv
 * @guid        adb5345337da4cd597840f79c8dc8fd8
 * @description Writes an MP3 frame into the output bitstream, adding
 *              the PlusV frames
 * @param       stream {mp3Stream_t, in} MP3 decoder's stream
 * @param       out {FILE*, in} The output file
 * @param       bufmp3 {unsigned char*, in} Buffer containing the MP3 frame
 * @param       lenmp3 {unsigned, in} Length of the MP3 frame
 * @param       bufpv {unsigned char*, in} Buffer containing the +V frame(s)
 * @param       lenpv {unsigned, in} Length of the +V frame(s)
 * @param       reserve {unsigned, in} Reserve space left over from the
 *              previous MP3 frame
 * @return      {unsigned} The reserve for the next writemp3pv call.
 * @note        bufmp3 only contains the bitstream data for the MP3 frame.
 *              Header and side information are stored in stream.
 */
unsigned writemp3pv( mp3Stream_t stream, FILE* out, unsigned char* bufmp3,
                     unsigned lenmp3, unsigned char* bufpv, unsigned lenpv,
                     unsigned reserve )
{
    mp3State_t* str = (mp3State_t*)stream;
    unsigned required;
    unsigned framelen;
    unsigned bitrate;
    unsigned padding;
    unsigned char hdrbuf[4];
    unsigned char mdbbuf;
    unsigned brtbl;
    unsigned brfrq;
    unsigned origrsv = reserve;

    /* Fill the reserve */

    if( reserve > lenmp3 )
    {
        /* All the MP3 data fits to the previous frame's reserve */

        fwrite( bufmp3, 1, lenmp3, out );
        reserve -= lenmp3;
        lenmp3 = 0;

        if( reserve > lenpv )
        {
            /* Also all +V data goes there */

            fwrite( bufpv, 1, lenpv, out );
            reserve -= lenpv;
            lenpv = 0;

            /* Fill the rest with a nice filler byte.
               This wastes space, but will only occur when wasting space
               is usually necessary anyway (when the bitstream continuously
               requires less than eight kbits/s of data) */

            for( ; reserve > 0; reserve-- )
            {
                fputc( 0x55, out );
            }
        }
        else
        {
            /* Store as much of the +V data as possible */

            fwrite( bufpv, 1, reserve, out );
            bufpv += reserve;
            lenpv -= reserve;
            reserve = 0;
        }
    }
    else
    {
        /* Store as much of the MP3 bitstream as possible */

        fwrite( bufmp3, 1, reserve, out );
        bufmp3 += reserve;
        lenmp3 -= reserve;
        reserve = 0;
    }

    /* Calculate the size of the new MP3 frame:
       4 for the MP3 header, plus
       side information as in the original MP3 frame, plus
       the portion of MP3 bitstream that did not fit to the reserve, plus
       the portion of +V bitstream that did not fit to the reserve. */

    required = 4 + str->_sidelength + lenmp3 + lenpv;

    /* Find the best match for the frame's bit rate. The first one that
       has at least the needed number of bytes is selected; the byte reserve
       is kept as small as possible. */

    brtbl = str->info.mpegVersion >= 20 ? 1 : 0;
    brfrq = str->info.mpegVersion >= 20 ? 2 * str->info.frequency
                                        : str->info.frequency;

    for( bitrate = 1; bitrate < 15; bitrate++ )
    {
        framelen = 144 * bitrates[brtbl][3][bitrate] * 1000 / brfrq;
        if( framelen + 1 >= required )
        {
            break;
        }
    }

    if( bitrate >= 15 )
    {
        /* Not enough space! Select the largest bitrate and enable padding. */

        bitrate = 14;
        framelen = 144 * bitrates[brtbl][3][bitrate] * 1000 /
                   brfrq + 1;
        padding = 1;

        required -= lenpv;
        lenpv = 0;
        if( framelen < required )
        {
            /* There is no space even for the MP3 bits! A very, very
               pathological case, but theoretically possible. Store as many
               MP3 bits as possible, but drop the rest. */

            lenmp3 -= required - framelen;
            required = framelen;
            fprintf( stderr, "WARNING! Dropping MP3 bits (too big bitrate)\n" );
        }
        else
        {
            /* The MP3 bitstream can be left intact, but we have to drop the
               +V information. This is not fatal, though, so do it. */

            fprintf( stderr, "Warning! Dropping +V bits (too big bitrate)\n" );
        }
    }
    else
    if( framelen + 1 == required )
    {
        padding = 1;
        framelen++;
    }
    else
    {
        padding = 0;
    }

    /* Update the statistics counters */

    bruse[bitrate]++;

    /* Edit the header */

    hdrbuf[0] = str->header[0];
    hdrbuf[1] = str->header[1] | 0x01;      /* no crc */
    hdrbuf[2] = ( str->header[2] & 0xd ) |
                ( bitrate << 4 ) |          /* new bitrate index */
                ( padding ? 2 : 0 );        /* padding */
    hdrbuf[3] = str->header[3];

    /* Write the header and side information */

    fwrite( hdrbuf, 1, 4, out );
    if( str->info.mpegVersion >= 20 )
    {
        mdbbuf = origrsv;
        fwrite( &mdbbuf, 1, 1, out );
        fwrite( str->_sideinfo + 1, 1, str->_sidelength - 1, out );
    }
    else
    {
        mdbbuf = origrsv >> 1;
        fwrite( &mdbbuf, 1, 1, out );
        mdbbuf = ( str->_sideinfo[1] & 0x7f ) | (( origrsv & 1 ) << 7 );
        fwrite( &mdbbuf, 1, 1, out );
        fwrite( str->_sideinfo + 2, 1, str->_sidelength - 2, out );
    }

    /* Write MP3 data */

    if( lenmp3 > 0 )
    {
        fwrite( bufmp3, 1, lenmp3, out );
    }

    /* Write +V data */

    if( lenpv > 0 )
    {
        fwrite( bufpv, 1, lenpv, out );
    }

    /* Calculate the new reserve */

    return framelen - required;
}

/**
 * @function    flushmp3pv
 * @guid        f833c06af04545ffac5af6ad1e69c824
 * @description Flushes the reserve for the last MP3 frame of the stream.
 * @param       out {FILE*, in} Output file name
 * @param       reserve {unsigned, in} Reserve size
 */
void flushmp3pv( FILE* out, unsigned reserve )
{
    for( ; reserve > 0; reserve-- )
    {
        fputc( 0x55, out );
    }
}

/**
 * @function    interleave
 * @guid        b410d72e533d44c2a0c4dfe6176b19a7
 * @description Interleaving function. Called by the lame+v encoder
 *              as well as the main function in this file.
 * @param       inmp3 {FILE*, in} Input MP3 stream
 * @param       inpv {FILE*, in} Input +V stream
 * @param       out {FILE*, in} Output MP3 stream
 */
void interleave( FILE* inmp3, FILE* inpv, FILE* out )
{
    mp3Stream_t mp3str;
    unsigned mp3reserve = 0;
    unsigned i;
    unsigned largest, digits;
    unsigned frames = 0;
    unsigned total;
    unsigned brtbl;

    /* Zero statistics */

    for( i = 0; i < 16; i++ )
    {
        bruse[i] = 0;
    }

    /* Open the MP3 decoder's stream */

    mp3InitDecoder( malloc, free );
    mp3str = mp3OpenStream( (void*)request, (void*)inmp3, NULL, NULL );

    while( !feof( inmp3 ) && !feof( inpv ))
    {
        unsigned char bufpv[4096];
        unsigned char bufmp3[4096];
        unsigned lenpv = 0;
        unsigned pvc = 0;
        unsigned pvcount;
        unsigned lenmp3;

        /* Decode a frame of MP3 data, and store the bits in the bufmp3
           buffer */

        bitsStartCopy( &((mp3State_t*)mp3str)->bits, bufmp3, 4096 );
        if( mp3DecodeFrame( mp3str ) != MP3_OK )
        {
            break;
        }
        lenmp3 = bitsEndCopy( &((mp3State_t*)mp3str)->bits );

        /* Read as many +V frames as needed for the MP3 frame */

        do
        {
            lenpv += readpv( inpv, bufpv + lenpv, &pvcount );
            pvc++;
        }
        while( pvc < pvcount );

        if( lenpv == 0 )
        {
            break;
        }

        /* Write out a new MP3 frame */

        mp3reserve = writemp3pv( mp3str, out, bufmp3, lenmp3,
                                 bufpv, lenpv, mp3reserve );

        frames++;
    }

    /* Flush the output */

    flushmp3pv( out, mp3reserve );

    /* Display statistics */

    largest = 0;
    total = 0;
    brtbl = ((mp3State_t*)mp3str)->info.mpegVersion >= 20 ? 1 : 0;
    for( i = 1; i < 15; i++ )
    {
        if( largest < bruse[i] )
        {
            largest = bruse[i];
        }
        total += bruse[i] * bitrates[brtbl][3][i];
    }
    digits = 1;
    if( largest >= 10 ) digits++;
    if( largest >= 100 ) digits++;
    if( largest >= 1000 ) digits++;
    if( largest >= 10000 ) digits++;
    if( largest >= 100000 ) digits++;
    if( largest >= 1000000 ) digits++;

    for( i = 1; i < 15; i++ )
    {
        char gauge[80];
        unsigned len = ( 80 - 8 - digits ) * bruse[i] / largest;

        memset( gauge, '*', len );
        gauge[len] = 0;

        printf( "%3i [%*i] %s\n", bitrates[brtbl][3][i], digits,
                bruse[i], gauge );
    }

    printf( "average: %.1f kbps\n", (double)total / (double)frames );
}

#ifndef NO_MAIN

/**
 * @function    main
 * @guid        1567738a93f846bf96ccff2a5fe1d341
 * @description A main() function that can be used to interleave MP3 and
 *              +V bitstreams available in files
 * @param       argc {int, in} Argument count
 * @param       argv[] {char*, in} Arguments
 * @return      {int} Status
 */
/*int main( int argc, char* argv[] )
{
    int i;
    int help = 1;

    FILE* inmp3 = 0;
    FILE* inpv = 0;
    FILE* out = 0;

    for( i = 1; i < argc; i++ )
    {
        if( argv[i][0] == '-' && argv[i][1] != 0 )
        {
            switch( argv[i][1] )
            {
                case 'h':
                case 'H':
                case '?':
                    help = 1;
                    i = argc;
                    break;
            }
        }
        else
        {
            int stdio = argv[i][0] == '-';

            if( inmp3 == 0 )
            {
                if( stdio )
                {
                    inmp3 = stdin;
                }
                else
                {
                    inmp3 = fopen( argv[i], "rb" );
                    if( inmp3 == 0 )
                    {
                        fprintf( stderr, "Unable to open %s for reading\n",
                                 argv[i] );
                        return 1;
                    }
                }
            }
            else
            if( inpv == 0 )
            {
                if( stdio )
                {
                    if( inmp3 == stdin )
                    {
                        fprintf( stderr, "MP3 and +V inputs cannot "
                                 "be stdin simultaneously\n" );
                        return 1;
                    }
                    inpv = stdin;
                }
                else
                {
                    inpv = fopen( argv[i], "rb" );
                    if( inpv == 0 )
                    {
                        fprintf( stderr, "Unable to open %s for reading\n",
                                 argv[i] );
                        return 1;
                    }
                }
            }
            else
            if( out == 0 )
            {
                if( stdio )
                {
                    out = stdout;
                }
                else
                {
                    out = fopen( argv[i], "wb" );
                    if( out == 0 )
                    {
                        fprintf( stderr, "Unable to open %s for writing\n",
                                 argv[i] );
                    }

                    help = 0;
                }
            }
            else
            {
                help = 1;
                break;
            }
        }
    }

    if( help )
    {
        fprintf( stderr, "%s [in/mp3] [in/+v] [out/mp3+v]\n", argv[0] );
        return 1;
    }

    interleave( inmp3, inpv, out );

    fclose( inmp3 );
    fclose( inpv );
    fclose( out );
    return 0;
}*/

#endif

