/*

Copyright 2001 VLSI Solution Oy.

This file is part of PlusV Reference Software.

PlusV Reference Software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

PlusV Reference Software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PlusV Reference Software; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
/*

  bitstream.c

  This file contains functions for encoding and decoding a PlusV bitstream.

  (C) 2001 VLSI Solution OY, http://www.vlsi.fi/
  Author: Henrik Herranen

  This reference source code may be used freely as long as the PlusV format
  is not intentionally broken!

  */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "bitstream.h"



static void BWriteBits(u_int8 data, int bits);
#if 0
static void BWriteBit(u_int8 data);
#endif
static u_int8 BReadBit(void);
static u_int8 BReadBits(int bits);



struct PlusVEncode {
  int low, high;
  int prefixBits, prefix;
  int codeBits, offset;
};

struct PlusVBitEncode {
  int values;
  struct PlusVEncode encode[10];
};

static const struct PlusVBitEncode bEncode[3] = {
  {
    48,
    {
      {   0,   1-1, 1, 1, 1,     0 },
      {  -1,  -0-1, 1, 1, 1,  1+ 1 },
      {   1,   2-1, 2, 1, 1, -1    },
      {  -2,  -1-1, 2, 1, 1,  2+ 1 },
      {   2,   4-1, 3, 1, 2, -2    },
      {  -4,  -2-1, 3, 1, 2,  4+ 2 },
      {   4,   8-1, 4, 1, 3, -4    },
      {  -8,  -4-1, 4, 1, 3,  8+ 4 },
      {   8,  24-1, 4, 0, 5, -8    },
      { -24,  -8-1, 4, 0, 5, 24+16 },
    }
  }, {
    96,
    {
      {   0,   2-1, 1, 1, 2,      0 },
      {  -2,  -0-1, 1, 1, 2,   2+ 2 },
      {   2,   4-1, 2, 1, 2,  -2    },
      {  -4,  -2-1, 2, 1, 2,   4+ 2 },
      {   4,   8-1, 3, 1, 3,  -4    },
      {  -8,  -4-1, 3, 1, 3,   8+ 4 },
      {   8,  16-1, 4, 1, 4,  -8    },
      { -16,  -8-1, 4, 1, 4,  16+ 8 },
      {  16,  48-1, 4, 0, 6, -16    },
      { -48, -16-1, 4, 0, 6,  48+32 },
    }
  }, {
    192,
    {
      {   0,   4-1, 1, 1, 3,      0 },
      {  -4,  -0-1, 1, 1, 3,   4+ 4 },
      {   4,   8-1, 2, 1, 3,  -4    },
      {  -8,  -4-1, 2, 1, 3,   8+ 4 },
      {   8,  16-1, 3, 1, 4,  -8    },
      { -16,  -8-1, 3, 1, 4,  16+ 8 },
      {  16,  32-1, 4, 1, 5, -16    },
      { -32, -16-1, 4, 1, 5,  32+16 },
      {  32,  96-1, 4, 0, 7, -32    },
      { -96, -32-1, 4, 0, 7,  96+64 },
    }
  }
};

struct PlusVCode {
  int prefixBits, prefixVal;
  int codeBits, posOffset, negOffset;
};

struct PlusVBitDecode {
  int values;
  struct PlusVCode code[5];
};

static const struct PlusVBitDecode bDecode[3] = {
  {
    48,
    {
      { 1, 1, 1, 0,  -1- 1},
      { 2, 1, 1, 1,  -2- 1},
      { 3, 1, 2, 2,  -4- 2},
      { 4, 1, 3, 4,  -8- 4},
      { 4, 0, 5, 8, -24-16}
    }
  }, {
    96,
    {
      { 1, 1, 2,  0,  -2- 2},
      { 2, 1, 2,  2,  -4- 2},
      { 3, 1, 3,  4,  -8- 4},
      { 4, 1, 4,  8, -16- 8},
      { 4, 0, 6, 16, -48-32}
    }
  }, {
    48,
    {
      { 1, 1, 3,  0,  -4- 4},
      { 2, 1, 3,  4,  -8- 4},
      { 3, 1, 4,  8, -16- 8},
      { 4, 1, 5, 16, -32-16},
      { 4, 0, 7, 32, -96-64}
    }
  }
};



/*

  BDecodeNumber() decodes a number from a PlusV bitstream.

  Inputs:
    int table - Quantisizing table to use

  Output:
    int - decoded value

 */
static int BDecodeNumber(int table) {
  int bits = 0, i = 0, res = 0, quit = 0;

  /* Traverse bit tree */
  while (!quit) {
    if (bits < bDecode[table].code[i].prefixBits) {
      bits++;
      res = (res << 1) | BReadBit();
    } else if (res != bDecode[table].code[i].prefixVal) {
      i++;
    } else {
      quit = 1;
    }
  }

  bits = bDecode[table].code[i].codeBits;
  res = BReadBits(bits);

  if (res >= (1<<(bits-1)))
    res += bDecode[table].code[i].negOffset;
  else
    res += bDecode[table].code[i].posOffset;

  return res;
}



/*

  BEncodeNumber() encodes a number to a PlusV bitstream.

  Inputs:
    int table - Quantisizing table to use
    int n - The value to be encoded. If the number is out of range, it is
	modulated with the maximum allowed number by the table.

  Output:
    int - decoded value

 */
static void BEncodeNumber(int table, int n) {
  int i = 0;

  /* Normalize to the allowed range */
  while (n >= (bEncode[table].values>>1))
    n -= bEncode[table].values;
  while (n < -(bEncode[table].values>>1))
    n += bEncode[table].values;

  /* Find correct range */
  while (n < bEncode[table].encode[i].low || n > bEncode[table].encode[i].high)
    i++;

  BWriteBits(bEncode[table].encode[i].prefix,
	    bEncode[table].encode[i].prefixBits);
  BWriteBits(n + bEncode[table].encode[i].offset,
	    bEncode[table].encode[i].codeBits);
}


static u_int16 *bitstream, *bitstreamStart;
static int bitShift = 16;
static u_int8 bitMask[9] = {
  0, 0x1, 0x3, 0x7, 0xf, 0x1f, 0x3f, 0x7f, 0xff
};



/*

  BBufInit() initializes a bit buffer of u_int16's for usage as a PlusV
  bitstream. Note that the table is intentionally of type u_int16 and not for
  instance char. This doesn't make any difference for PC based systems, but it
  makes the algorithm more suitable for 16-bit DSPs.

  Inputs:
    u_int16 *bstream - A pointer to the bit buffer to be initialized. For a
	valid PlusV block, 213 bits is the maximum possible size in the
	format, i.e. an allocated space of 14 words is enough for the table.

  Output:
    none

 */
static void BBufInit(u_int16 *bstream) {
  bitstream = bitstreamStart = bstream;
  bitShift = 16;
}



/*

  BWriteBits() writes bits to an initialized PlusV bit buffer.

  Inputs:
    u_int8 data - The data to be written.
    int bits - The number of bits to write (1..8). It is worth noting that
	this function can access only 8 bits of data although some PlusV
	fields are be upto 11 bits long!

  Output:
    none

 */
static void BWriteBits(u_int8 data, int bits) {
  bitShift -= bits;
  if (bitShift < 0) {
    u_int16 tmpD = *bitstream << 8;
    tmpD |= (data & bitMask[bits]) << (bitShift+8);
    *bitstream++ |= tmpD>>8;
    *bitstream |= tmpD<<8;
    bitShift += 16;
  } else {
    *bitstream |= (data & bitMask[bits]) << bitShift;
    if (!bitShift) {
      bitstream++;
      bitShift += 16;
    }
  }
}



#if 0
/*

  BWriteBit() writes a bit to an initialized PlusV bit buffer.

  Inputs:
    u_int8 data - The data to be written. The LSB of the data is used as the
	bit to be written.

  Output:
    none

 */
static void BWriteBit(u_int8 data) {
  bitShift--;
  if (data & 1) {
    *bitstream |= 1<<bitShift;
  }
  if (!bitShift) {
    bitShift = 16;
    bitstream++;
  }
}
#endif



/*

  BPos() return the bit position in the current PlusV bit buffer.

  Inputs:
    none

  Output:
    The current position in the table. As usual in C, the base index is 0.

 */
int BPos(void) {
  return 16*(bitstream-bitstreamStart) + (16-bitShift);
}



/*

  ByteWrapPutC() writes a byte to a file. The original intent of this function
  is that if there is a need to escape some byte codes, this function acts as
  a wrapper for them. However, for MP3 this just writes all bytes as is.

  Inputs:
    FILE *fp - The file to write the byte to.
    int c - The byte to write.

  Output:
    The character written on success. If an error occurs, EOF is returned.

 */
int ByteWrapPutC(FILE *fp, int c) {
  return putc(c, fp);
}



/*

  ByteWrapGetC() reads a byte from a file. The original intent of this
  function is that if there is a need to escape some byte codes, this
  function acts as an unwrapper for them. However, for MP3 this just reads
  all bytes as is.

  Inputs:
    FILE *fp - The file to read the byte from.

  Output:
    The character read on success. If an error occurs, EOF is returned.

 */
int ByteWrapGetC(FILE *fp) {
  return fgetc(fp);
}



/*

  BSaveBBuf() writes the current PlusV bit buffer to a file.

  Inputs:
    FILE *fp - The file to write the bit buffer to.

  Output:
    The number of bytes successfully written, or negative on error.

 */
static int BSaveBBuf(FILE *fp) {
  int bytes = (BPos()+7) >> 3, i, wrote = 0;
  const u_int16 *localBitstream = bitstreamStart;

  for (i=0; i<bytes>>1; i++) {
    if (ByteWrapPutC(fp, *localBitstream   >> 8) == EOF)
      return wrote;
    wrote++;
    if (ByteWrapPutC(fp, *localBitstream++ & 0xff) == EOF)
      return wrote;
    wrote++;
  }
  if (bytes & 1)
    if (ByteWrapPutC(fp, *localBitstream   >> 8) != EOF)
      wrote++;

  return wrote;
}



/*

  BLoadBBuf() initialized and loads a PlusV bit buffer from a file.

  Inputs:
    FILE *fp - The file to read the bit buffer from.
    u_int16 *bbuf - Pointer to a bit buffer.
    int bytes - The number of bytes to read.

  Output:
    The number of bytes successfully read, or negative on error.

 */
static int BLoadBBuf(FILE *fp, u_int16 *bbuf, int bytes) {
  int i;
  BBufInit(bbuf);
  for (i=0; i<bytes>>1; i++) {
    u_int16 t = ByteWrapGetC(fp) << 8;
    *bbuf++ = t | ByteWrapGetC(fp);
  }
  if (bytes & 1) {
    *bbuf = ByteWrapGetC(fp) << 8;
  }
  return (feof(fp) ? -1 : bytes);
}



/*

  BReadBit() reads a bit from a PlusV bit buffer.

  Inputs:
    none

  Output:
    The resulting bit in the LSB of the return value.

 */
static u_int8 BReadBit(void) {
  u_int8 res = (*bitstream >> --bitShift) & 1;
  if (!bitShift) {
    bitShift = 16;
    bitstream++;
  }
  return res;
}



/*

  BReadBits() reads bits from a PlusV bit buffer.

  Inputs:
    int bits - The number of bits to be read (1..8).

  Output:
    The resulting bits.

 */
static u_int8 BReadBits(int bits) {
  u_int8 res = 0;
  do {
    res = (res << 1) | BReadBit();
  } while (--bits);
  return res;
}



/*

  RoundEnergy() quantizes an energy value using the given quantization table.

  Inputs:
    flotari energy - Floating point energy value to be quantized.
    int quant - Quantisizing table to use

  Output:
    The resulting, quantized value. Negative value is returned for error.

 */
int RoundEnergy(flotari energy, int quant) {
  flotari fQuant = 0;

  switch(quant) {
  case 0:
    fQuant = 0.5;
    break;
  case 1:
    fQuant = 1.0;
    break;
  case 2:
    fQuant = 2.0;
    break;
  default:
    return -1;
  }

  return floor(energy * fQuant + 0.5);
}



#if 0
/*

  Test() is an internal test function. Do not touch.

 */
static void Test(void) {
  u_int16 pafi[100] = {0};
  int i;

  BBufInit(pafi);
  for (i=-0; i<24; i++)
    BEncodeNumber(0, i);

  for (i=0; i<8; i++)
    fprintf(stderr, "0x%04x ", pafi[i]);
  fprintf(stderr, "-> %d bits\n", BPos());

  BBufInit(pafi);
  for (i=-0; i<24; i++) {
    fprintf(stderr, "%d->%d\n", i, BDecodeNumber(0));
  }

  exit(EXIT_FAILURE);
}
#endif



/*

  WriteFatPlusVBlock() writes the given PlusV structure as a Fat Data Block.

  Inputs:
    FILE *fp - File pointer to write the file to.
    struct PlusV *v - PlusV structure with all fields set to valid values.

  Output:
    0 for success, negative for error

 */
int WriteFatPlusVBlock(FILE *fp, struct PlusV *v) {
  int i, ch;
  u_int16 bBuf[100] = {0};

  BBufInit(bBuf);

  switch(v->quant) {
  case 0:
    v->oldQuant = 1;
    break;
  case 1:
    v->oldQuant = 2;
    break;
  case 2:
    v->oldQuant = 4;
    break;
  default:
    fprintf(stderr, "Quantization value is wrong: %d (should be 1..3)\n",
	    v->quant);
    return -1;
  }

  for (i=0; i<4; i++) {
    if (v->sines[i]) {
      v->oldSineFreq[i] = v->sineFreq[i];
    } else {
      v->oldSineFreq[i] = 0;
    }
    v->oldSines[i] = v->sines[i];
  }

  for (ch=0; ch<v->channels; ch++) {
    for (i=0; i<4; i++) {
      if (v->sines[i]) {
	if (v->sineEnergy[ch][i] < bEncode[v->quant].values)
	  v->oldSineEnergy[ch][i] = v->sineEnergy[ch][i];
	else
	  v->oldSineEnergy[ch][i] = bEncode[v->quant].values - 1;
      } else {
	v->oldSineEnergy[ch][i] = 0;
      }
    }

    for (i=0; i<8; i++) {
      if (v->noiseEnergy[ch][i] < bEncode[v->quant].values)
	v->oldNoiseEnergy[ch][i] = v->noiseEnergy[ch][i];
      else
	v->oldNoiseEnergy[ch][i] = bEncode[v->quant].values - 1;
    }
  }

  BWriteBits(0xd6, 8);
  BWriteBits(0x53, 8);
  BWriteBits(0xd6, 8);
  BWriteBits(0xfd, 8);

  BWriteBits(v->version, 2);
  BWriteBits(0, 6); /* v->blockLen unknown at this time */
  BWriteBits(v->bands, 3);
  BWriteBits(v->quant, 2);
  BWriteBits(v->channels, 3);
  BWriteBits(v->private, 2);
  BWriteBits(v->trans, 1);
  if (v->trans)
    BWriteBits(v->trPos, 2);

  {
    int t=0;
    for (i=0; i<4; i++)
      t = (t<<1) | (v->sines[i] ? 1 : 0);
    BWriteBits(t, 4);
  }
  for (i=0; i<4; i++) {
    if (v->sines[i]) {
#if 0
      fprintf(stderr, "sF[%d]=%3d\n", i, v->oldSineFreq[i]);
#endif
      BWriteBits(v->oldSineFreq[i], 7);
    }
  }

  for (ch=0; ch<v->channels; ch++) {

    for (i=0; i<4; i++) {
      if (v->sines[i]) {
	BWriteBits(v->oldSineEnergy[ch][i], 6+v->quant);
      }
    }
    for (i=0; i<v->bands; i++) {
      BWriteBits(v->oldNoiseEnergy[ch][i], 6+v->quant);
    }
  }

  bBuf[2] |= ((BPos()+7)>>3) << 8;
  BSaveBBuf(fp);

  v->counter = 1;

  return 0;
}



/*

  WriteLeanPlusVBlock() writes the given PlusV structure as a Lean Data Block.

  Inputs:
    FILE *fp - File pointer to write the file to.
    struct PlusV *v - PlusV structure with all fields set to valid values.

  Output:
    0 for success, negative for error

 */
int WriteLeanPlusVBlock(FILE *fp, struct PlusV *v) {
  int i, ch;
  u_int16 bBuf[100] = {0};

  BBufInit(bBuf);

#if 0
  Test();
#endif

  for (ch=0; ch<v->channels; ch++) {
    for (i=0; i<4; i++) {
      if (v->sineEnergy[ch][i] >= bEncode[v->quant].values)
	v->sineEnergy[ch][i] = bEncode[v->quant].values - 1;
    }

    for (i=0; i<8; i++) {
      if (v->noiseEnergy[ch][i] >= bEncode[v->quant].values)
	v->noiseEnergy[ch][i] = bEncode[v->quant].values - 1;
    }
  }

  BWriteBits(v->counter, 2);
  BWriteBits(0, 6); /* v->blockLen unknown at this time */
  BWriteBits(v->trans, 1);
  if (v->trans)
    BWriteBits(v->trPos, 2);


  /* Collect which sines are in use */
  {
    int t=0;

    for (i=0; i<4; i++)
      t = (t<<1) | (v->sines[i] ? 1 : 0);
    BWriteBits(t, 4);
  }

  /* Write sinusoids */
  for (i=0; i<4; i++) {
    if (v->sines[i]) {
      int t = v->sineFreq[i] - v->oldSineFreq[i];
      if (t >= (bEncode[2].values >> 1))
	t -= 128;
      else if (t < -(bEncode[2].values >> 1))
	t += 128;
      BEncodeNumber(2, t);
      /* Encoding is always successful */
      v->oldSineFreq[i] = v->sineFreq[i];
    }
    v->oldSines[i] = v->sines[i];
  }

  for (ch=0; ch<v->channels; ch++) {
    /* Write sinusoids */
    for (i=0; i<4; i++) {
      if (v->sines[i]) {
	BEncodeNumber(v->quant, v->sineEnergy[ch][i]-v->oldSineEnergy[ch][i]);
	v->oldSineEnergy[ch][i] = v->sineEnergy[ch][i];
      } else {
	v->oldSineEnergy[ch][i] = 0;
      }
    }

    /* Write noise energies */
    for (i=0; i<v->bands; i++) {
      BEncodeNumber(v->quant, v->noiseEnergy[ch][i]-v->oldNoiseEnergy[ch][i]);
      v->oldNoiseEnergy[ch][i] = v->noiseEnergy[ch][i];
    }
  }

  bBuf[0] |= ((BPos()+7)>>3) << 8;
  BSaveBBuf(fp);

  if (++v->counter == 3)
    v->counter = 0;

  return 0;
}



/*

  WritePlusVBlock() writes the given PlusV structure to a file.

  Inputs:
    FILE *fp - File pointer to write the file to.
    struct PlusV *v - PlusV structure with all fields set to valid values.
    int fat - If non-zero, write the block as a Fat Data Block, otherwise
	as a Lean Data Block.

  Output:
    0 for success, negative for error

 */
int WritePlusVBlock(FILE *fp, struct PlusV *v, int fat) {
  if (fat)
    return WriteFatPlusVBlock(fp, v);
  else
    return WriteLeanPlusVBlock(fp, v);
}





/*

  ReadFatPlusVBlock() reads a Fat Data Block. This function assumes that
  the first four bytes of the FDB, i.e. field fHeader has already been read.

  Inputs:
    FILE *fp - File pointer to read the file from.
    struct PlusV *v - PlusV structure to which the values read are to be
	written to.

  Output:
    0 for success, negative for error

 */
static int ReadFatPlusVBlock(FILE *fp, struct PlusV *v) {
  int i, ch;
  u_int16 bBuf[64] = {0};

  BLoadBBuf(fp, bBuf, 3);
  if (BReadBits(8) != 0x53 || BReadBits(8) != 0xd6 || BReadBits(8) != 0xfd)
    return -2;

  BLoadBBuf(fp, bBuf, 1);
  v->version = BReadBits(2);
  v->blockLen = BReadBits(6);
  BLoadBBuf(fp, bBuf, v->blockLen-5);
  v->bands = BReadBits(3);
  v->quant = BReadBits(2);
  v->oldQuant = (!v->quant) ? 4 : ((v->quant == 1) ? 2 : 1);
  v->channels = BReadBits(3);
  v->private = BReadBits(2);
  v->trans = BReadBits(1);
  if (v->trans)
    v->trPos = BReadBits(2);

  {
    int t = BReadBits(4);
    for (i=3; i>=0; i--) {
      v->sines[i] = t&1;
      t >>= 1;
    }
  }

  for (i=0; i<4; i++) {
    if (v->sines[i]) {
      v->sineFreq[i] = BReadBits(7);
    } else {
      v->sineFreq[i] = 0;
    }
  }

  for (ch=0; ch<v->channels; ch++) {
    for (i=0; i<4; i++) {
      if (v->sines[i]) {
	v->sineEnergy[ch][i] = BReadBits(6+v->quant) * v->oldQuant;
      } else {
	v->sineEnergy[ch][i] = 0;
      }
    }


    for (i=0; i<v->bands; i++)
      v->noiseEnergy[ch][i] = BReadBits(6+v->quant) * v->oldQuant;
    for (i=v->bands; i<8; i++)
      v->noiseEnergy[ch][i] = 0;
#if 0
    /* Debug */
    for (i=0; i<8; i++)
      v->noiseEnergy[ch][i] = 0;
#if 1
    v->noiseEnergy[ch][0] = 188;
#endif
#endif


#if 0
    /* Debug */
    for (i=0; i<4; i++) {
      v->sines[i] = 0;
      v->sineFreq[i] = 0;
      v->sineEnergy[ch][i] = 0;
    }
#if 0
    v->sines[0] = 1;
    v->sineFreq[0] = 16;
    v->sineEnergy[ch][0] = 188;
#endif
#endif
  }

  v->counter = 1;

  return 0;
}



/*

  ReadLeanPlusVBlock() reads a Lean Data Block. This function assumes that
  the first byte, i.e. fields lCounter and lBLockLen has already been read
  and checked for.

  Inputs:
    FILE *fp - File pointer to read the file from.
    struct PlusV *v - PlusV structure to which the values read are to be
	written to.
    int length - The contents of field lBlockLen.

  Output:
    0 for success, negative for error

 */
static int ReadLeanPlusVBlock(FILE *fp, struct PlusV *v, int length) {
  int i, ch;
  u_int16 bBuf[64] = {0};

  BLoadBBuf(fp, bBuf, length-1);
  v->trans = BReadBits(1);
  if (v->trans)
    v->trPos = BReadBits(2);

  {
    int t = BReadBits(4);
    for (i=3; i>=0; i--) {
      v->sines[i] = t&1;
      t >>= 1;
    }
  }

  /* Read sinusoids */
  for (i=0; i<4; i++) {
    if (v->sines[i]) {
      v->sineFreq[i] += BDecodeNumber(2);
      /* Wrapping is better done with "&= 127" than with "%= 128" */
      v->sineFreq[i] &= 127;
    }
  }

  for (ch=0; ch<v->channels; ch++) {
    /* Read sinusoids */
    for (i=0; i<4; i++) {
      if (v->sines[i]) {
	int t;
	t = BDecodeNumber(v->quant) * v->oldQuant + v->sineEnergy[ch][i];
	while (t >= 192)
	  t -= 192;
	while (t < 0)
	  t += 192;
	v->sineEnergy[ch][i] = t;
      } else {
	v->sineEnergy[ch][i] = 0;
      }
    }

    /* Read noise energies */
    for (i=0; i<v->bands; i++) {
      int t = BDecodeNumber(v->quant) * v->oldQuant + v->noiseEnergy[ch][i];
      while (t >= 192)
	t -= 192;
      while (t < 0)
	t += 192;
      v->noiseEnergy[ch][i] = t;
    }
    for (i=v->bands; i<8; i++)
      v->noiseEnergy[ch][i] = 0;

  }

  if (++v->counter == 3)
    v->counter = 0;

  return 0;
}



/*

  ReadPlusVBlock() reads a PlusV data block, decides if it's a valid Fat or
  Lean Data Block, and decodes it to the PlusV structure.

  Inputs:
    FILE *fp - File pointer to read the file from.
    struct PlusV *v - PlusV structure to which the values read are to be
	written to.

  Output:
    0 for success, -1 for read error, -2 for data error.

 */
int ReadPlusVBlock(FILE *fp, struct PlusV *v) {
  int c;
  u_int16 bBuf[1] = {0};

  BLoadBBuf(fp, bBuf, 1);
  if ((c = BReadBits(8)) != 0xd6) {
#if 0
    fprintf(stderr, "Counter = %d, Header = 0x%2x\n", v->counter, c);
#endif
    if (c>>6 == v->counter)
      return ReadLeanPlusVBlock(fp, v, c&0x3f);
    else
      return -2;
  }
  return ReadFatPlusVBlock(fp, v);
}
