/*

Copyright 2001 Ionific Ltd.

This file is part of PlusV Reference Software.

PlusV Reference Software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

PlusV Reference Software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PlusV Reference Software; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
/**
 * @file        fft0512.c
 * @guid        6366d5c8b6034d5a8c5feeca93a6a9fb
 * @copyright   2001 Ionific Ltd.
 * @author      Jussi Lahdenniemi, jussi.lahdenniemi@ionific.com
 * @project     plusv/encoder
 * @description 512-long real-to-complex FFT
 */

#include <math.h>
#include "fft0512.h"

#ifndef M_PI
#   define M_PI 3.14159265358979323846264
#endif

/* Length of FFT: number of components / log2 number of components */

#define N 512
#define LOGN 9

/**
 * @function    fft512
 * @guid        f22dd77d5c0b4f87a06bf292e079648f
 * @description Performs the FFT.
 * @param       x[] {double, in}
 *              Real-valued input vector; complex-valued output is also
 *              placed on this vector.
 *              x[0] is the real part of FFT's output [0]
 *              x[1] is the real part of FFT's output [N/2]
 *              x[2] is the real part of FFT's output [1]
 *              x[3] is the imaginary part of FFT's output [1]
 *              x[4] is the real part of FFT's output [2]
 *              x[5] is the imaginary part of FFT's output [2]
 *              x[6] is the real part of FFT's output [3]
 *              x[7] is the imaginary part of FFT's output [3]
 *              ...
 *              FFT's output components [0] and [N/2] always have a zero
 *              imaginary part, so they are packed together as shown above.
 * @note        This implementation is not terribly efficient:
 *              - Transform could be done in-place
 *              - A N/2-long FFT is enough to transform N pure real values.
 */
void fft512( flotari x[] )
{
    static int tableinit = 0;
    static flotari costbl[N];
    static flotari sintbl[N];

    flotari re1[N], im1[N];
    flotari re2[N], im2[N];
    flotari *res, *ims, *red, *imd;
    int i;
    int ia, ib, ic, id, m, k;

    /* Initialize (co)sine tables */

    if( !tableinit )
    {
        tableinit = 1;

        for( i = 0; i < N; i++ )
        {
            costbl[i] = cos( -(flotari)i * 2. * M_PI / (flotari)N );
            sintbl[i] = sin( -(flotari)i * 2. * M_PI / (flotari)N );
        }
    }

    /* Copy to complex vector */

    for( i = 0; i < N; i++ )
    {
        re1[i] = x[i];
        im1[i] = 0.;
    }

    ia = N / 2;
    ib = 1;

    res = re1;
    ims = im1;
    red = re2;
    imd = im2;

    /* Complex-valued FFT */

    for ( i = 0; i < LOGN; i++ )
    {
        ic = 0;
        id = ia;
        for ( k = 1; k <= ib; k++ )
        {
            flotari zre = costbl[ic];
            flotari zim = sintbl[ic];

            for( m = ic; m < id; m++ )
            {
                flotari are = res[m+ic];
                flotari aim = ims[m+ic];
                flotari bre = zre * res[m+id] - zim * ims[m+id];
                flotari bim = zim * res[m+id] + zre * ims[m+id];

                red[m] = are + bre;
                imd[m] = aim + bim;
                red[m+N/2] = are - bre;
                imd[m+N/2] = aim - bim;
            }
            ic = id;
            id += ia;
        }
        ia /= 2;
        ib *= 2;

        if( res == re1 )
        {
            res = re2;
            ims = im2;
            red = re1;
            imd = im1;
        }
        else
        {
            res = re1;
            ims = im1;
            red = re2;
            imd = im2;
        }
    }

    /* Copy the result back to x */

    for( i = 0; i < N/2; i++ )
    {
        x[i*2] = res[i];
        x[i*2+1] = ims[i];
    }

    x[1] = res[N/2];
}

