/*

Copyright 1995-2001 Henrik Herranen.

This file is part of PlusV Reference Software.

PlusV Reference Software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

PlusV Reference Software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PlusV Reference Software; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "leo_types.h"
#include "leo_endian.h"
#include "leo_sound.h"


static struct SAMPLE_HEADER sampleHeader;

static lbool ReadRawByteHeader(u_int32 magic);
static lbool ReadRawShortHeader(u_int32 magic);
static lbool ReadRawShortLeHeader(u_int32 magic);
static lbool Read8SVXHeader(u_int32 magic);
static lbool ReadAuHeader(u_int32 magic);
static lbool ReadWavHeader(u_int32 magic);
static lbool ReadLB16Header(u_int32 magic);
static s_int32 ReadUnsignedRawByteSample(void);
static s_int32 ReadSignedRawByteSample(void);
static s_int32 ReadUnsignedRawShortSample(void);
static s_int32 ReadSignedRawShortSample(void);
static s_int32 Read8SVXSample(void);
static s_int32 ReadAuUSample(void);
static s_int32 ReadAuWSample(void);
static s_int32 ReadRawUlawSample(void);
static s_int32 ReadWavBSample(void);
static s_int32 ReadWavWSample(void);
static s_int32 ReadWavLSample(void);
static s_int32 ReadWavUSample(void);
static s_int32 ReadLB16Sample(void);
static s_int32 ReadSignedRawLeSample(void);


static lbool (*headerFuncs[F_UNKNOWN])(u_int32) = {
        ReadRawByteHeader,
        ReadRawByteHeader,
        ReadRawShortHeader,
        ReadRawShortHeader,
        Read8SVXHeader,
        ReadAuHeader,
	ReadAuHeader,
	ReadRawByteHeader,
	ReadWavHeader,
	ReadWavHeader,
	ReadWavHeader,
	ReadWavHeader,
	ReadLB16Header,
	ReadRawShortLeHeader
};

static s_int32 (*inFuncs[F_UNKNOWN])(void) = {
        ReadUnsignedRawByteSample,
        ReadSignedRawByteSample,
        ReadUnsignedRawShortSample,
        ReadSignedRawShortSample,
        Read8SVXSample,
        ReadAuUSample,
	ReadAuWSample,
        ReadRawUlawSample,
	ReadWavUSample,
	ReadWavBSample,
	ReadWavWSample,
	ReadWavLSample,
	ReadLB16Sample,
	ReadSignedRawLeSample
};
static int funcNo = 2;
static lbool (*headerFunc)(u_int32);
static s_int32 (*inFunc)(void);
static FILE *inFile;
static lbool magicStored;
static u_int32 magicStore;

struct SAMPLE_HEADER *ReadSampleHeader(FILE *fp) {
    u_int32 magic;
    int lastChar;

    magic = (u_int32)fgetc(fp) << 24;
    magic |= (u_int32)fgetc(fp) << 16;
    magic |= (u_int32)fgetc(fp) << 8;
    lastChar = fgetc(fp);
    if (lastChar == EOF) {
	fprintf(stderr, "Couldn't read sample header magic number, "
			"file error\n");
	return NULL;
    }
    return ReadSampleHeaderMagic(fp, F_UNKNOWN, magic|(u_int32)lastChar);
}

struct SAMPLE_HEADER *ReadSampleHeaderMagic(FILE *fp,
						enum SND_FILETYPE fileType,
						u_int32 magic) {
    if (fileType >= F_UNKNOWN) {
	if (magic == ID_RIFF) {
	    fileType = F_WAVB;
	} else if (magic == ID_FORM) {
	    fileType = F_8SVX;
	} else if (magic == ID_SND) {
	    fileType = F_AUU;
	} else if ((BigEndian() ? magic : U32toBE(magic)) == ID_LB16) {
	    fileType = F_LB16;
	} else {
	    fprintf(stderr, "Cannot read unknown file type (%08x)\n",
		    fileType);
	    return NULL;
	}
    }
    funcNo = fileType;
    headerFunc = headerFuncs[fileType];
    inFunc = inFuncs[fileType];
    inFile = fp;

    memset(&sampleHeader, 0, sizeof(sampleHeader));

    if (!headerFunc(magic)) {
	fprintf(stderr, "Couldn't read sample header.\n");
	return NULL;
    }

    sampleHeader.fileType = fileType;

    return &sampleHeader;
}





u_int32 ReadSamples(s_int16 *pos, u_int32 n) {
    u_int32 nRead = 0xFFFFFFFF;
    int c;

    while (++nRead < n && (c = inFunc()) != LEOF) {
	*pos++ = (short)(c>>16);
    }

    return nRead;
}


u_int32 ReadLSamples(s_int32 *pos, u_int32 n) {
    u_int32 nRead = 0xFFFFFFFF;
    int c;

    while (++nRead < n && (c = inFunc()) != LEOF) {
	*pos++ = (long)c;
    }

    return nRead;
}





static lbool ReadRawByteHeader(u_int32 magic) {
    magicStored = 4;
    magicStore = magic;
    sampleHeader.channels   =  1;
    sampleHeader.sampleRate =  0;
    sampleHeader.bitsPerSample = (funcNo == F_ULAW) ? -8 : 8;
    if (inFile == stdin || fseek(inFile, 0, SEEK_END) ||
	(sampleHeader.samples = ftell(inFile)) < 0 ||
	fseek(inFile, 0, SEEK_SET)) {
	fprintf(stderr, "Warning: sample size couldn't be determined\n");
	sampleHeader.samples    =  -1;
    }
    return TRUE;
}

static lbool ReadRawShortHeader(u_int32 magic) {
    magicStored = 4;
    magicStore = magic;
    sampleHeader.channels   =  1;
    sampleHeader.sampleRate =  0;
    sampleHeader.bitsPerSample = 16;
    if (inFile == stdin || fseek(inFile, 0, SEEK_END) ||
	(sampleHeader.samples = ftell(inFile)) < 0 ||
	fseek(inFile, 0, SEEK_SET)) {
	fprintf(stderr, "Warning: sample size couldn't be determined\n");
	sampleHeader.samples    =  -1;
    }
    return TRUE;
}

static lbool ReadRawShortLeHeader(u_int32 magic) {
    magicStored = 4;
    magicStore = magic;
    sampleHeader.channels   =  2;
    sampleHeader.sampleRate =  0;
    sampleHeader.bitsPerSample = 16;
    if (inFile == stdin || fseek(inFile, 0, SEEK_END) ||
	(sampleHeader.samples = ftell(inFile)) < 0 ||
	fseek(inFile, 0, SEEK_SET)) {
	fprintf(stderr, "Warning: sample size couldn't be determined\n");
	sampleHeader.samples    =  -1;
    }
    return TRUE;
}

static s_int32 len8svx;

static lbool Read8SVXHeader(u_int32 magic) {
    u_int8 iffh[8];
    struct IFF_8SVX_HEADER head;
    struct IFF_8SVX_HEADER_VOICE8CHAN chan;
    s_int32 left;
    lbool foundVHDR = FALSE, foundBODY = FALSE;
    u_int32 i, s;

    magicStored = 0;

    if (magic != ID_FORM) {
	fprintf(stderr, "Not an IFF file.\n");
	return FALSE;
    }

    if (fread(iffh, sizeof(iffh), 1, inFile) != 1)
	return FALSE;

    if (MakeID(iffh[4], iffh[5], iffh[6], iffh[7]) != ID_8SVX) {
	fprintf(stderr, "Not an IFF 8SVX file.\n");
	return FALSE;
    }

    left = MakeID(iffh[0], iffh[1], iffh[2], iffh[3]);

    while(left > 0 && (foundBODY == FALSE || foundVHDR == FALSE)) {
	if (fread(&i, sizeof(i), 1, inFile) != 1)
	    return FALSE;
	if (fread(&s, sizeof(s), 1, inFile) != 1)
	    return FALSE;
	i = BEtoU32(i);
	s = BEtoU32(s);
	if (i == ID_VHDR) {
	    if (s != sizeof(head)) {
		fprintf(stderr, "Unexpected header size " FMT_U32 ", "
			"expected %d.\n",
			s, (int)sizeof(head));
		return FALSE;
	    }
	    if (fread(&head, sizeof(head), 1, inFile) != 1)
		return FALSE;

	    sampleHeader.channels = 1;
	    sampleHeader.sampleRate = BEtoU16(head.samplesPerSec);
	    len8svx = BEtoU32(head.oneShotHiSamples) +
		BEtoU32(head.repeatHiSamples);
	    sampleHeader.samples = len8svx;
	    sampleHeader.bitsPerSample = 8;
	    foundVHDR = TRUE;
	} else if (i == ID_CHAN) {
	    if (s != sizeof(chan)) {
		fprintf(stderr, "Unexpected CHAN size " FMT_U32 ", "
			"expected %d.\n", s, (int)sizeof(chan));
		return FALSE;
	    }
	    if (fread(&chan, sizeof(chan), 1, inFile) != 1)
		return FALSE;
	    switch (BEtoS32(chan.chanCode)) {
	    case CHAN_MONO_RIGHT:
	    case CHAN_MONO_LEFT:
		sampleHeader.channels = 1;
		break;
	    case CHAN_STEREO:
		sampleHeader.channels = 2;
		break;
	    default:
		fprintf(stderr, "Warning: unrecognized CHAN code " FMT_U32 ", "
			"ignoring\n", BEtoU32(chan.chanCode));
		break;
	    }
	} else if (i == ID_BODY) {
	    if (!foundVHDR) {
		fprintf(stderr, "Can't read 8SVX with BODY before VHDR.\n");
		return FALSE;
	    }
	    if (s != len8svx) {
		fprintf(stderr,
			"VHDR claimed body size " FMT_S32 ", BODY claims "
			FMT_U32 "\n",
			len8svx, s);
		return FALSE;

	    }
	    foundBODY = TRUE;
	} else {
	    int skip = s;
	    while (skip--)
		if (fgetc(inFile) == EOF) {
			fprintf(stderr, "Unexpected end of 8SVX file.\n");
			return FALSE;
		}
	}
	left -= (8+s);
    }

    sampleHeader.samples /= sampleHeader.channels;

    return TRUE;
}

static s_int32 lenAu = -1;

static lbool ReadAuHeader(u_int32 magic) {
    struct SUN_AU_HEADER head;
    long i;

    magicStored = 0;

    if (magic != ID_SND) {
	fprintf(stderr, "Not an AU file.\n");
	return FALSE;
    }

    if (fread(((char *)&head)+4, sizeof(head)-4, 1, inFile) != 1)
	return FALSE;

    for (i=0; i<BEtoU32(head.headerSize)-sizeof(head); i++)
	if (fgetc(inFile) == EOF)
	    return FALSE;

    sampleHeader.sampleRate = BEtoU32(head.sampleRate);
    sampleHeader.channels = BEtoU32(head.channels);
    sampleHeader.samples = BEtoU32(head.dataSize) / sampleHeader.channels;
    if (BEtoU32(head.encoding) != 1) {
	if (BEtoU32(head.encoding) != 3) {
	    fprintf(stderr, "Not an 8-bit ulaw nor 16-bit linear au.\n");
	    return FALSE;
	} else {
	    sampleHeader.bitsPerSample = 16;
	    sampleHeader.samples /= 2;
	    inFunc = ReadAuWSample;
	}
    } else {
	    sampleHeader.bitsPerSample = -8;
	    inFunc = ReadAuUSample;
    }
    if (sampleHeader.channels < 1) {
	fprintf(stderr, "Can't handle " FMT_S32 " channel samples.\n",
		sampleHeader.channels);
	return FALSE;
    }
    lenAu = sampleHeader.channels * sampleHeader.samples;
    if (lenAu < 1) {
	lenAu = -1;
    }

    return TRUE;
}


static s_int32 lenWav = -1;

static lbool ReadWavHeader(u_int32 magic) {
    u_int8 h[8];
    struct RIFF_WAV_HEADER fmt;
    s_int32 left;
    lbool foundFmt = FALSE, foundData = FALSE;
    u_int32 i, s;

    magicStored = 0;

    if (magic != ID_RIFF) {
	fprintf(stderr, "Not a RIFF file.\n");
	return FALSE;
    }

    if (fread(h, sizeof(h), 1, inFile) != 1)
	return FALSE;

    if (MakeID(h[4], h[5], h[6], h[7]) != ID_WAVE) {
	fprintf(stderr, "Not a RIFF WAVE (.wav) file.\n");
	return FALSE;
    }

    left = MakeID(h[3], h[2], h[1], h[0]);

    while(left > 0 && (foundData == FALSE || foundFmt == FALSE)) {
	if (fread(&i, sizeof(i), 1, inFile) != 1)
	    return FALSE;
	if (fread(&s, sizeof(s), 1, inFile) != 1)
	    return FALSE;
	i = BEtoU32(i);
	s = LEtoU32(s);
	if (i == ID__FMT) {
	    if (s < sizeof(fmt)) {
		fprintf(stderr, "Unexpected header size " FMT_U32 ", "
			"expected %d.\n", s, (int)sizeof(fmt));
		return FALSE;
	    }
	    if (fread(&fmt, sizeof(fmt), 1, inFile) != 1)
		return FALSE;
	    for (i=0; i<s-sizeof(fmt); i++)
		fgetc(inFile);


	    switch(LEtoU16(fmt.formatTag)) {
	    case WAVE_FORMAT_PCM:
		switch(LEtoU16(fmt.bitsPerSample)) {
		case 8:
		    inFunc = ReadWavBSample;
		    sampleHeader.bitsPerSample = 8;
		    break;
		case 16:
		    inFunc = ReadWavWSample;
		    sampleHeader.bitsPerSample = 16;
		    break;
		case 32:
		    inFunc = ReadWavLSample;
		    sampleHeader.bitsPerSample = 32;
		    break;
		default:
		    fprintf(stderr, "Can't read %d bit RIFF WAVE PCM samples "
			    "(only 8, 16 and 32 bits supported)\n",
			    LEtoU16(fmt.bitsPerSample));
		    return FALSE;
		}
		break;
	    case WAVE_FORMAT_MULAW:
		if (LEtoU16(fmt.bitsPerSample) == 8) {
		    inFunc = ReadWavUSample;
		    sampleHeader.bitsPerSample = -8;
		} else {
		    fprintf(stderr, "Can't read %d bit RIFF WAVE MULAW samples"
			    " (only 8 bits supported)\n",
			    LEtoU16(fmt.bitsPerSample));
		    return FALSE;
		}
		break;
	    default:
		fprintf(stderr, "RIFF WAVE format %d not supported\n",
			LEtoU16(fmt.formatTag));
		return FALSE;
		break;
	    }
	    sampleHeader.sampleRate = LEtoU32(fmt.samplesPerSecond);
	    sampleHeader.channels = LEtoU16(fmt.channels);
	    foundFmt = TRUE;
	} else if (i == ID_DATA) {
	    if (!foundFmt) {
		fprintf(stderr, "Can't read WAVE with DATA before FMT.\n");
		return FALSE;
	    }
	    lenWav = s * 8 / LEtoU16(fmt.bitsPerSample);
	    sampleHeader.samples = lenWav / sampleHeader.channels;
	    foundData = TRUE;
	} else {
	    int skip = s;
	    while (skip--)
		if (fgetc(inFile) == EOF) {
			fprintf(stderr, "Unexpected end of RIFF WAVE file.\n");
			return FALSE;
		}
	}
	left -= (8+s);
    }

    return TRUE;
}


static s_int32 lenLb = -1;

static lbool ReadLB16Header(u_int32 magic) {
    struct LB16_HEADER head;

    magicStored = 0;

    if ((BigEndian() ? magic : U32toBE(magic)) != ID_LB16) {
	fprintf(stderr, "Not a LB16 file.\n");
	return FALSE;
    }

    if (fread(((char *)&head)+4, sizeof(head)-4, 1, inFile) != 1)
	return FALSE;

    sampleHeader.samples = head.samples / head.channels;
    sampleHeader.sampleRate = head.sampleRate;
    sampleHeader.channels = head.channels;
    sampleHeader.bitsPerSample = 16;

    if (sampleHeader.channels < 1) {
	fprintf(stderr, "Can't handle " FMT_S32 " channel samples.\n",
		sampleHeader.channels);
	return FALSE;
    }

    if (sampleHeader.samples < 1)
	lenLb = -1;
    else
	lenLb = sampleHeader.channels * sampleHeader.samples;

    return TRUE;
}





/******* RAW SAMPLE READ FUNCTIONS ********/

static long RU(void) {
    int c;
    unsigned char ulawbyte;
    static int exp_lut[8] = { 0, 132, 396, 924, 1980, 4092, 8316, 16764 };
    int sign, exponent, mantissa, sample;
    if (magicStored) {
	c = (magicStore >> 24) & 0xFF;
	magicStore <<= 8;
	magicStored--;
    } else {
	c = getc(inFile);
    }
    ulawbyte = c;

    if (!lenAu || c == EOF)
	return LEOF;

    ulawbyte = ~ ulawbyte;
    sign = ( ulawbyte & 0x80 );
    exponent = ( ulawbyte >> 4 ) & 0x07;
    mantissa = ulawbyte & 0x0F;
    sample = exp_lut[exponent] + ( mantissa << ( exponent + 3 ) );
    if ( sign != 0 )
	sample = -sample;

    return sample << 16;
}

static long RUB(void) {
    int c;
    if (magicStored) {
	c = (magicStore >> 24) & 0xFF;
	magicStore <<= 8;
	magicStored--;
    } else {
	c = getc(inFile);
    }
    return (c == EOF) ? LEOF : (((c<<8)+c)-32768) << 16;
}

static long RSB(void) {
    int c;
    if (magicStored) {
	c = (magicStore >> 24) & 0xFF;
	magicStore <<= 8;
	magicStored--;
    } else {
	c = getc(inFile);
    }
    return (c == EOF) ? LEOF : ((c^0x80)*0x101-0x8000) << 16;
}

static long RUW(void) {
    int c1 = getc(inFile), c2 = getc(inFile);
    return (c1 == EOF || c2 == EOF) ? LEOF : ((c1<<8) + c2 - 32768) << 16;
}


static long RSW(void) {
    int c1 = getc(inFile), c2 = getc(inFile);
    return (c1 == EOF || c2 == EOF) ? LEOF :
	((((c1<<8) + c2) ^ 0x8000) - 32768) << 16;
}

static long RSWLE(void) {
    int c1 = getc(inFile), c2 = getc(inFile);
    return (c1 == EOF || c2 == EOF) ? LEOF :
	((((c2<<8) + c1) ^ 0x8000) - 32768) << 16;
}

static long RSLLE(void) {
    int c1 = getc(inFile), c2 = getc(inFile),
	c3 = getc(inFile), c4 = getc(inFile);
    long res;
    if (c4 == EOF)
	return LEOF;

    res =
	((c2<<8) + c1) +
	((((c4<<8) + c3) ^ 0x8000) << 16);

    if (res == LEOF)
	res--;

    return res;
}





/******* HIGHER LEVEL SAMPLE READ FUNCTIONS ********/

static s_int32 ReadUnsignedRawByteSample(void) {
    return RUB();
}

static s_int32 ReadSignedRawByteSample(void) {
    return RSB();
}

static s_int32 ReadUnsignedRawShortSample(void) {
    return RUW();
}


static s_int32 ReadSignedRawShortSample(void) {
    return RSW();
}

static s_int32 ReadSignedRawLeSample(void) {
    return RSWLE();
}



static s_int32 Read8SVXSample() {
    if (!len8svx)
	return LEOF;
    len8svx--;
    return RSB();
}

static s_int32 ReadAuUSample() {
    if (!lenAu)
	return LEOF;
    lenAu--;
    return RU();
}

static s_int32 ReadRawUlawSample() {
    return RU();
}

static s_int32 ReadAuWSample() {
    if (!lenAu)
	return LEOF;
    lenAu--;
    return RSW();
}

static s_int32 ReadWavUSample() {
    if (!lenWav)
	return LEOF;
    lenWav--;
    return RU();
}

static s_int32 ReadWavBSample() {
    if (!lenWav)
	return LEOF;
    lenWav--;
    return RUB();
}

static s_int32 ReadWavWSample() {
    if (!lenWav)
	return LEOF;
    lenWav--;
    return RSWLE();
}

static s_int32 ReadWavLSample() {
    if (!lenWav)
	return LEOF;
    lenWav--;
    return RSLLE();
}

static s_int32 ReadLB16Sample() {
    s_int16 t;
    if (!lenLb || fread(&t, sizeof(s_int16), 1, inFile) != 1)
	return LEOF;

    lenLb--;
    return ((s_int32)t)<<16;
}
