/*

Copyright 1995-2001 Henrik Herranen.

This file is part of PlusV Reference Software.

PlusV Reference Software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

PlusV Reference Software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PlusV Reference Software; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include "leo_types.h"
#include "leo_endian.h"
#include "leo_sound.h"


static struct SAMPLE_HEADER sampleHeader;

/*struct BUFFER {
    long size;
    long sampleRate;
    unsigned char *data;
};*/


static lbool WriteRawUHeader(enum SND_FILETYPE fileType);
static lbool WriteRawBHeader(enum SND_FILETYPE fileType);
static lbool WriteRawWHeader(enum SND_FILETYPE fileType);
static lbool Write8SVXHeader(enum SND_FILETYPE fileType);
static lbool WriteAuHeader(enum SND_FILETYPE fileType);
static lbool WriteWavHeader(enum SND_FILETYPE fileType);
static lbool WriteLB16Header(enum SND_FILETYPE fileType);

static lbool WriteUnsignedRawByteSample(int c);
static lbool WriteSignedRawByteSample(int c);
static lbool WriteUnsignedRawShortSample(int c);
static lbool WriteSignedRawShortSample(int c);
static lbool Write8SVXSample(int c);
static lbool WriteAuUSample(int c);
static lbool WriteAuWSample(int c);
static lbool WriteWavUSample(int c);
static lbool WriteWavBSample(int c);
static lbool WriteWavWSample(int c);
static lbool WriteWavLSample(int c);
static lbool WriteLB16Sample(int c);

static lbool WriteRawFooter(void);
static lbool Write8SVXFooter(void);
static lbool WriteAuFooter(void);
static lbool WriteWavFooter(void);
static lbool WriteLB16Footer(void);


int (*headerFuncs[F_UNKNOWN])(enum SND_FILETYPE) = {
    WriteRawBHeader,
    WriteRawBHeader,
    WriteRawWHeader,
    WriteRawWHeader,
    Write8SVXHeader,
    WriteAuHeader,
    WriteAuHeader,
    WriteRawUHeader,
    WriteWavHeader,
    WriteWavHeader,
    WriteWavHeader,
    WriteWavHeader,
    WriteLB16Header
};
int (*outFuncs[F_UNKNOWN])(int) = {
    WriteUnsignedRawByteSample,
    WriteSignedRawByteSample,
    WriteUnsignedRawShortSample,
    WriteSignedRawShortSample,
    Write8SVXSample,
    WriteAuUSample,
    WriteAuWSample,
    WriteAuUSample,
    WriteWavUSample,
    WriteWavBSample,
    WriteWavWSample,
    WriteWavLSample,
    WriteLB16Sample
};
int (*footerFuncs[F_UNKNOWN])(void) = {
    WriteRawFooter,
    WriteRawFooter,
    WriteRawFooter,
    WriteRawFooter,
    Write8SVXFooter,
    WriteAuFooter,
    WriteAuFooter,
    WriteRawFooter,
    WriteWavFooter,
    WriteWavFooter,
    WriteWavFooter,
    WriteWavFooter,
    WriteLB16Footer
};

static lbool (*headerFunc)(enum SND_FILETYPE);
static lbool (*outFunc)(int);
static lbool (*footerFunc)(void);
static FILE *fp;
static long left;

struct SAMPLE_HEADER *WriteSampleHeader(FILE *fiP, enum SND_FILETYPE fileType,
				s_int32 channels, s_int32 sampleRate,
				s_int32 samples) {
    headerFunc = headerFuncs[fileType];
    outFunc = outFuncs[fileType];
    footerFunc = footerFuncs[fileType];
    fp = fiP;

    if (channels < 1) {
	fprintf(stderr, FMT_S32 " is not a proper number of channels\n",
		channels);
	return NULL;
    }
    sampleHeader.channels = channels;
    sampleHeader.sampleRate = sampleRate;
    sampleHeader.samples = samples;
    sampleHeader.fileType = fileType;

    left = (sampleHeader.samples < 0) ? 0x80000000 : sampleHeader.samples *
	channels;

    if (!headerFunc(fileType)) {
	fprintf(stderr, "Couldn't write file header. Exiting.\n");
	exit(EXIT_FAILURE);
    }

    return &sampleHeader;
}

int WriteSampleFooter(void) {
    if (!footerFunc()) {
	fprintf(stderr, "Couldn't write file footer. Exiting.\n");
	exit(EXIT_FAILURE);
    }

    if (left != 0x80000000) {
	if (left > 0) {
	    fprintf(stderr, "Warning: left %ld samples unwritten\n", left);
	} else if (left < 0) {
	    fprintf(stderr, "Warning: wrote %ld samples too much\n", -left);
	}
    }
    return TRUE;
}

lbool WriteSamples(s_int16 *pos, u_int32 n) {
    while(n--) {
	if (left!=0x80000000)
	    left--;
	if (!outFunc(((s_int32) *pos++) << 16))
	    return FALSE;
    }

    return TRUE;
}


lbool WriteLSamples(s_int32 *pos, u_int32 n) {
    while(n--) {
	if (left!=0x80000000)
	    left--;
	if (!outFunc(*pos++))
	    return FALSE;
    }

    return TRUE;
}

/****** ACTUAL FUNCTIONS */

static lbool WriteRawUHeader(enum SND_FILETYPE fileType) {
    sampleHeader.bitsPerSample = -8;
    return TRUE;
}

static lbool WriteRawBHeader(enum SND_FILETYPE fileType) {
    sampleHeader.bitsPerSample = 8;
    return TRUE;
}

static lbool WriteRawWHeader(enum SND_FILETYPE fileType) {
    sampleHeader.bitsPerSample = 16;
    return TRUE;
}


static u_int32 svxLen, svxAllOverhead;
static u_int32 svxAllPos, svxBodyPos, svxVhdrPos;

static lbool Write8SVXHeader(enum SND_FILETYPE fileType) {
    struct IFF_8SVX_HEADER head = {0};
    u_int32 ul;

    if (sampleHeader.channels < 1 || sampleHeader.channels > 1) {
	fprintf(stderr, "Can't encode " FMT_S32 " channels to IFF 8SVX\n",
		sampleHeader.channels);
	return FALSE;
    }
    if (sampleHeader.sampleRate > 65535 || sampleHeader.sampleRate < 0) {
	fprintf(stderr, "Can't encode sample rate " FMT_S32 " Hz to IFF 8SVX "
		"(must be between 0..65535 Hz)\n",
		sampleHeader.sampleRate);
	return FALSE;
    }
    svxAllOverhead = 4 + 8 + sizeof(head) + 8 + 0/*BODY size=0*/;

    sampleHeader.bitsPerSample = 8;

    head.oneShotHiSamples = U32toBE(sampleHeader.samples);
    head.samplesPerSec = U16toBE((u_int16)sampleHeader.sampleRate);
    head.ctOctave = 1;
    head.volume = S32toBE(0x10000);

    svxLen = 0;

    ul = U32toBE(ID_FORM);
    if (fwrite(&ul, 4, 1, fp) != 1)
	return FALSE;
    svxAllPos = ftell(fp);
    ul = U32toBE(svxAllOverhead + sampleHeader.samples);
    if (fwrite(&ul, 4, 1, fp) != 1)
	return FALSE;
    ul = U32toBE(ID_8SVX);
    if (fwrite(&ul, 4, 1, fp) != 1)
	return FALSE;
    ul = U32toBE(ID_VHDR);
    if (fwrite(&ul, 4, 1, fp) != 1)
	return FALSE;
    ul = U32toBE(sizeof(head));
    if (fwrite(&ul, 4, 1, fp) != 1)
	return FALSE;
    svxVhdrPos = ftell(fp);	/* oneShotHiSamples position */
    if (fwrite(&head, sizeof(head), 1, fp) != 1)
	return FALSE;
    ul = U32toBE(ID_BODY);
    if (fwrite(&ul, 4, 1, fp) != 1)
	return FALSE;
    svxBodyPos = ftell(fp);
    ul = U32toBE(sampleHeader.samples); /* Body size == 0 == unknown */
    if (fwrite(&ul, 4, 1, fp) != 1)
	return FALSE;

    return TRUE;
}


static u_int32 auLen = 0;

static lbool WriteAuHeader(enum SND_FILETYPE fileType) {
    struct SUN_AU_HEADER head = {{0}};

    if (sampleHeader.channels < 1) {
	fprintf(stderr, "Can't encode " FMT_S32 " channels to SUN AU\n",
		sampleHeader.channels);
	return FALSE;
    }
    if (sampleHeader.sampleRate < 0) {
	fprintf(stderr, "Can't encode sample rate " FMT_S32 " Hz to SUN AU "
		"(must be > 0 Hz)\n",
		sampleHeader.sampleRate);
	return FALSE;
    }

    auLen = 0;

    sampleHeader.bitsPerSample = (fileType == F_AUU) ? -8 : 16;

    head.magic[0] = '.';
    head.magic[1] = 's';
    head.magic[2] = 'n';
    head.magic[3] = 'd';
    head.headerSize = U32toBE(sizeof(head));
    head.dataSize = U32toBE(sampleHeader.samples *
			      ((fileType == F_AUW) ? 2 : 1) *
			      sampleHeader.channels);
    head.encoding = U32toBE((u_int32)((fileType == F_AUU) ? 1 : 3));
    head.sampleRate = U32toBE(sampleHeader.sampleRate);
    head.channels = U32toBE(sampleHeader.channels);
    if (fwrite(&head, sizeof(head), 1, fp) != 1)
	return FALSE;

    return TRUE;
}



static u_int32 wavLen, wavAllOverhead;
static u_int32 wavAllPos, wavDataPos;

static lbool WriteWavHeader(enum SND_FILETYPE fileType) {
    struct RIFF_WAV_HEADER fmt = {0};
    unsigned long ul;

    if (sampleHeader.channels < 1) {
	fprintf(stderr, "Can't encode " FMT_S32 " channels to RIFF WAVE\n",
		sampleHeader.channels);
	return FALSE;
    }
    if (sampleHeader.sampleRate < 0) {
	fprintf(stderr, "Can't encode sample rate " FMT_S32 " Hz to RIFF WAVE "
		"(must be > 0 Hz)\n",
		sampleHeader.sampleRate);
	return FALSE;
    }
    wavAllOverhead = 4 + 8 + sizeof(fmt) + 8 + 0/*DATA size=0*/;

    if (fileType == F_WAVU) {
	fmt.formatTag = U16toLE(WAVE_FORMAT_MULAW);
	sampleHeader.bitsPerSample = -8;
    } else if (fileType == F_WAVB) {
	fmt.formatTag = U16toLE(WAVE_FORMAT_PCM);
	sampleHeader.bitsPerSample = 8;
    } else if (fileType == F_WAVW) {
	fmt.formatTag = U16toLE(WAVE_FORMAT_PCM);
	sampleHeader.bitsPerSample = 16;
    } else {
	fmt.formatTag = U16toLE(WAVE_FORMAT_PCM);
	sampleHeader.bitsPerSample = 32;
    }
    fmt.channels = U16toLE(sampleHeader.channels);
    fmt.samplesPerSecond = U32toLE(sampleHeader.sampleRate);
    fmt.avgBytesPerSec = U32toLE(sampleHeader.channels *
				  sampleHeader.sampleRate *
				  abs(sampleHeader.bitsPerSample) / 8);
    fmt.blockAlign = U16toLE(sampleHeader.channels *
			       abs(sampleHeader.bitsPerSample) / 8);
    fmt.bitsPerSample = U16toLE(abs(sampleHeader.bitsPerSample));

    wavLen = 0;

    ul = U32toBE(ID_RIFF);
    if (fwrite(&ul, 4, 1, fp) != 1)
	return FALSE;
    wavAllPos = ftell(fp);
    ul = U32toLE(wavAllOverhead + sampleHeader.samples*sampleHeader.channels*
	(abs(sampleHeader.bitsPerSample)/8));
    if (fwrite(&ul, 4, 1, fp) != 1)
	return FALSE;
    ul = U32toBE(ID_WAVE);
    if (fwrite(&ul, 4, 1, fp) != 1)
	return FALSE;
    ul = U32toBE(ID__FMT);
    if (fwrite(&ul, 4, 1, fp) != 1)
	return FALSE;
    ul = U32toLE(sizeof(fmt));
    if (fwrite(&ul, 4, 1, fp) != 1)
	return FALSE;
    if (fwrite(&fmt, sizeof(fmt), 1, fp) != 1)
	return FALSE;
    ul = U32toBE(ID_DATA);
    if (fwrite(&ul, 4, 1, fp) != 1)
	return FALSE;
    wavDataPos = ftell(fp);
    ul = U32toLE(sampleHeader.samples*sampleHeader.channels*
	(abs(sampleHeader.bitsPerSample)/8)); /* Body size == 0 == unknown */
    if (fwrite(&ul, 4, 1, fp) != 1)
	return FALSE;

    return TRUE;
}




static u_int32 lb16Len = 0;

static lbool WriteLB16Header(enum SND_FILETYPE fileType) {
    struct LB16_HEADER head = {ID_LB16};

    if (sampleHeader.channels < 1) {
	fprintf(stderr, "Can't encode " FMT_S32 " channels to LB16\n",
		sampleHeader.channels);
	return FALSE;
    }
    if (sampleHeader.sampleRate < 0) {
	fprintf(stderr, "Can't encode sample rate " FMT_S32 " Hz to LB16 "
		"(must be >= 0 Hz)\n",
		sampleHeader.sampleRate);
	return FALSE;
    }

    lb16Len = 0;

    head.samples = sampleHeader.samples * sampleHeader.channels;
    head.sampleRate = sampleHeader.sampleRate;
    head.channels = sampleHeader.channels;
    if (fwrite(&head, sizeof(head), 1, fp) != 1)
	return FALSE;

    return TRUE;
}




/********** RAW WRITE FUNCTIONS **********/

#define ZEROTRAP    /* turn on the trap as per the MIL-STD */
#define BIAS 0x84   /* define the add-in bias for 16 bit samples */
#define CLIP 32635

static lbool WU(int c) {
    static int exp_lut[256] = {0,0,1,1,2,2,2,2,3,3,3,3,3,3,3,3,
                                   4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
                                   5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
                                   5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
                                   6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
                                   6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
                                   6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
                                   6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
                                   7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                                   7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                                   7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                                   7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                                   7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                                   7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                                   7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                                   7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7};
    int sign, exponent, mantissa;
    unsigned char ulawbyte;

    c = (c >> 16) & 0xFFFF;

    /* Get the sample into sign-magnitude. */
    sign = (c >> 8) & 0x80;                     /* set aside the sign */
    if ( sign != 0 ) c = -c;                    /* get magnitude */

    if ( c > CLIP ) c = CLIP;                   /* clip the magnitude */

    /* Convert from 16 bit linear to ulaw. */
    c += BIAS;
    exponent = exp_lut[( c >> 7 ) & 0xFF];
    mantissa = ( c >> ( exponent + 3 ) ) & 0x0F;
    ulawbyte = ~( sign | ( exponent << 4 ) | mantissa );
#ifdef ZEROTRAP
    if( ulawbyte == 0 )
        ulawbyte = 0x02;       /* optional CCITT trap */
#endif
    
    return (fputc(ulawbyte, fp)) == ulawbyte;
}


static lbool WSB(int c) {
    c = (((c+0x80000000) >> 24) ^ 0x80) & 0xFF;
    return putc(c, fp) == c;
}

static lbool WUB(int c) {
    c = ((c+0x80000000) >> 24) & 0xFF;
    return putc(c, fp) == c;
}

static lbool WSW(int c) {
    c = (((c + 0x80000000) >> 16) & 0xFFFF) ^ 0x8000;
    return putc(c>>8, fp) == c>>8 && putc(c&0xFF, fp) == (c&0xFF);
}

static lbool WUW(int c) {
    c = ((c + 0x80000000) >> 16) & 0xFFFF;
    return putc(c>>8, fp) == c>>8 && putc(c&0xFF, fp) == (c&0xFF);
}

static lbool WSWLE(int c) {
    c = (((c + 0x80000000) >> 16) & 0xFFFF) ^ 0x8000;
    return putc(c&0xFF, fp) == (c&0xFF) && putc(c>>8, fp) == c>>8;
}

static lbool WSLLE(int c) {
    int a[4];
    c ^= 0x80000000;
    a[0] = (c >> 24) & 0xFF;
    a[1] = (c >> 16) & 0xFF;
    a[2] = (c >>  8) & 0xFF;
    a[3] = (c      ) & 0xFF;
    return putc(a[3], fp) == a[3] && putc(a[2], fp) == a[2] &&
	putc(a[1], fp) == a[1] && putc(a[0], fp) == a[0];
}



/********** HIGHER LEVEL WRITE FUNCTIONS **********/


static lbool WriteSignedRawByteSample(int c) {
    return WSB(c);
}

static lbool Write8SVXSample(int c) {
    svxLen++;
    return WSB(c);
}

static lbool WriteUnsignedRawByteSample(int c) {
    return WUB(c);
}

static lbool WriteSignedRawShortSample(int c) {
    return WSW(c);
}

static lbool WriteUnsignedRawShortSample(int c) {
    return WUW(c);
}


static lbool WriteAuWSample(int c) {
    auLen += 2;
    return WSW(c);
}

static lbool WriteAuUSample(int c) {
    auLen++;
    return WU(c);
}

static lbool WriteWavUSample(int c) {
    wavLen++;
    return WU(c);
}

static lbool WriteWavBSample(int c) {
    wavLen++;
    return (WUB(c));
}

static lbool WriteWavWSample(int c) {
    wavLen+=2;
    return (WSWLE(c));
}

static lbool WriteWavLSample(int c) {
    wavLen+=4;
    return (WSLLE(c));
}

static lbool WriteLB16Sample(int c) {
    s_int16 t = c >> 16;
    lb16Len++;
    return (fwrite(&t, sizeof(s_int16), 1, fp) == 1);
}




/********** FOOTER WRITE FUNCTIONS **********/






static lbool WriteRawFooter(void) {
    return TRUE;
}

static lbool Write8SVXFooter(void) {
    if (svxLen & 1) {
	if (fputc(0, fp) != 0)
	    return FALSE;
	else
	    svxLen++;
    }

    if (sampleHeader.samples < 1) {
	sampleHeader.samples = svxLen;
	svxAllOverhead += svxLen;
	svxLen = U32toBE(svxLen);
	svxAllOverhead = U32toBE(svxAllOverhead);
	if (fp == stdout ||
	    fseek(fp, (long)svxAllPos, SEEK_SET) < 0 ||
	    fwrite(&svxAllOverhead, 4, 1, fp) != 1 ||
	    fseek(fp, (long)svxBodyPos, SEEK_SET) < 0 ||
	    fwrite(&svxLen, 4, 1, fp) != 1 ||
	    fseek(fp, (long)svxVhdrPos, SEEK_SET) < 0 ||
	    fwrite(&svxLen, 4, 1, fp) != 1) {
	    fprintf(stderr, "Couldn't write 8SVX size properly\n");
	    return FALSE;
	}
    }

    return TRUE;
}

static lbool WriteAuFooter(void) {
    if (sampleHeader.samples < 1) {
	sampleHeader.samples = auLen / (abs(sampleHeader.bitsPerSample / 8));
	auLen = U32toBE(auLen);
	if (fp == stdout ||
	    fseek(fp, 8, SEEK_SET) < 0 ||  /* 8 = length pos in au file */
	    fwrite(&auLen, 4, 1, fp) != 1) {
	    fprintf(stderr, "Couldn't write SUN AU size properly\n");
	    return FALSE;
	}
    }

    return TRUE;
}

static lbool WriteWavFooter(void) {
    if (wavLen & 1) {
	if (fputc(0, fp) != 0)
	    return FALSE;
	else
	    wavLen++;
    }

    wavAllOverhead += wavLen;
    if (sampleHeader.samples < 1) {
	sampleHeader.samples = wavLen / (abs(sampleHeader.bitsPerSample) / 8);
	wavAllOverhead = U32toLE(wavAllOverhead);
	wavLen = U32toLE(wavLen);
	if (fp == stdout ||
	    fseek(fp, (long)wavAllPos, SEEK_SET) < 0 ||
	    fwrite(&wavAllOverhead, 4, 1, fp) != 1 ||
	    fseek(fp, (long)wavDataPos, SEEK_SET) < 0 ||
	    fwrite(&wavLen, 4, 1, fp) != 1) {
	    fprintf(stderr, "Couldn't write RIFF WAV size properly\n");
	    return FALSE;
	}
    }

    return TRUE;
}

static lbool WriteLB16Footer(void) {
    if (sampleHeader.samples < 1) {
	if (fp == stdout ||
	    fseek(fp, 4, SEEK_SET) < 0 ||  /* 4 = samples pos in file */
	    fwrite(&lb16Len, 4, 1, fp) != 1) {
	    fprintf(stderr, "Couldn't write LB16 size properly\n");
	    return FALSE;
	}
    }

    return TRUE;
}
