/*

Copyright 2001 VLSI Solution Oy.

This file is part of PlusV Reference Software.

PlusV Reference Software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

PlusV Reference Software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PlusV Reference Software; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
/*

  plusv.c

  This is the main body for the PlusV encoder / decoder. It doesn't have
  anything MP3 specific details in it, except of hard coded encoder and
  decoder delays, which however can be changed easily.

  (C) 2001 VLSI Solution OY, http://www.vlsi.fi/
  Author: Henrik Herranen

  This reference source code may be used freely as long as the PlusV format
  is not intentionally broken!

  */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <malloc.h>	// for dynamic array allocation :) [gameplaya15143]
#include "leo_types.h"
#include "leo_sound.h"
#include "leo_endian.h"
#include "bitstream.h"
#include "filter.h"
#include "plusv.h"
#include "fft0512.h"
#include "pvrandom.h"

/* ENCODE_QUANT = 0..2, 0 = smallest, 2 = best */
//#define ENCODE_QUANT 0	// [gameplaya15143] commented out, see plusv.h
/* ENCODE_BANDS = 0..7 */
//[gameplay15143] commented this stuff out
//#define ENCODE_BANDS_48 5	/* 19.5 kHz bandwidth for 48   kHz */
//#define ENCODE_BANDS_44 6	/* 19.3 kHz bandwidth for 44.1 kHz */
//#define ENCODE_BANDS_32 7	/* 15.0 kHz bandwidth for 32   kHz */
/* ENCODE_VERSION = 0 */
#define ENCODE_VERSION 0

#if 1
/* Expanding has the potential of removing background noise when used with
   caution. */
#define EXPAND_SOUND
#define EXPAND_LIMIT 40.0
#define EXPAND_RATIO (2.0/1.0)
#endif

#if 0
#  define ALWAYS_ENCODE_SHORT_BLOCKS 0 /* Always encode shorter blocks */
#else
#  if 1
#    define ENCODE_SHORT_BLOCKS_AT_32K /* Short blocks only if fs = 32 kHz */
#  endif
#endif

/* 576 is the default MP3 decoder delay. */
#define MP3_DECODE_DELAY 576
/* 534 seems to be an approximately correct figure for LAME (guessed).
   Probably this should be parametrizable for different encoders. */
#define EXTERNAL_ENCODER_DELAY 534

//#define TOTAL_ENCODER_DELAY (MP3_DECODE_DELAY+EXTERNAL_ENCODER_DELAY)
//#define TOTAL_ENCODER_DELAY 1	//[gameplaya15143] commented out, see plusv.h
/* Encoding definitions */

#if 1
#define ENCODE_TRANSIENTS	/* Encode transient (f/lTrans = 1) blocks? */
#if 1
#define SHARPEN_TRANSIENTS	/* Try to make transient edges cleaner? */
#endif
#endif
/* The higher the following number, the less there are transient blocks */
#define TRANSIENT_ENERGY_UP_THRESHOLD 4.0	/* Energy*=4 -> transient */
#define TRANSIENT_ENERGY_DOWN_THRESHOLD 8.0	/* Energy/=4 -> transient */
#if 0
#define MARK_TRANSIENT_BLOCKS	/* Debug: mark audio samples */
#endif

#if 1
#define ENCODE_SINES	/* Encode also sines in addition to noise? */
/*
  The higher this threshold is, the less frequencies are encoded as sine
  waves instead of noise, and the smaller the file size gets.
*/
#define SIN_THRESHOLD 16.0
#endif

#if 0
#define ENCODE_ONLY_FAT_BLOCKS	/* Debug, creates larger, incompatible files */
#endif
#define FAT_BLOCK_LIMIT 66	/* How often do we need a fat block? */
	/* Note: 66 is divisible both by 2 and 3 so we get FDBs at beginning
	   of PlusV frames regardless of if use short or long blocks as long
	   as we don't change block length during the file. If we did,
	   there'd have to be a little more logic. */



/* Decode definitions */
#if 1
#define DECODE_UPPER	/* Debug: decode also PlusV data */
#if 1
#define DECODE_SINS	/* Debug: decode also PlusV sines */
#endif
#if 1
#define DECODE_NOISES	/* Debug: decode also PlusV noise */
#endif
#endif
#if 1
#define DECODE_LOWER	/* Debug: decode MP3 data */
#endif
#if 0
#define MARK_DECODE_BLOCK_LIMITS	/* Debug: mark audio samples */
#endif


/*
  The following correcting term is to correct dB scales. In our definition,
  we use the INCORRECT definition that 6 dB = 1/2, whereas the correct
  definition is that 20 dB = 1/10, which leads to that 1/2 is
  approximately 6.02 dB. This is done on purpose to make calculation easier.
  However, this also means that we have to compensate for that error from the
  specification document.
*/
#define DB_CORR (6.02059991327962/6.0)




#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#define max(x,y) ((x)>(y)?(x):(y))

/*
  Standard
  Blackman window WITH energy saving (= x/0.4192)
*/
void BlackmanWindow(flotari *fir, int taps) {
  int m;
  flotari sum = 0;
  for (m=0; m<taps; m++) {
    fir[m] = (.42 - .50*cos(2*M_PI*m/taps) + .08*cos(4*M_PI*m/taps));
    /* The next line makes energy to be saved even though Blackman is used */
    sum += fir[m];
  }
  sum /= m;
  for (m=0; m<taps; m++)
    fir[m] /= sum;
}







/* PlusV data structure needed by bitstream.c */
struct PlusV plusV = {0};


/*
  The following parameters are built-in MP3+V and cannot be changed as
  long as used with MP3.
*/
#define FFT_BUF_SIZE 512
#define FFT_SLICES 16
#define FFT_SLICE_SIZE (FFT_BUF_SIZE/FFT_SLICES/2)
#define PLUSV_SHORT_BLOCK_SIZE 384
#define PLUSV_LONG_BLOCK_SIZE 576



/*

  SwapSines() swaps two sine frequencies in a PlusV data structure.

  Inputs:
    struct PlusV *v - PlusV structure with all fields set to valid values.
    int x1 - Index of first sine to be swapped.
    int x2 - Index of second sine to be swapped.

  Output:
    none

 */
static void SwapSines(struct PlusV *v, int x1, int x2) {
    int ch, t;
    t = v->sines[x1];
    v->sines[x1] = v->sines[x2];
    v->sines[x2] = t;
    t = v->sineFreq[x1];
    v->sineFreq[x1] = v->sineFreq[x2];
    v->sineFreq[x2] = t;
    for (ch=0; ch<v->channels; ch++) {
      t = v->sineEnergy[ch][x1];
      v->sineEnergy[ch][x1] = v->sineEnergy[ch][x2];
      v->sineEnergy[ch][x2] = t;
    }
}




static flotari Expand(flotari f) {
#if 0
  fprintf(stderr, "f=%f\n", f);
#endif
#ifdef EXPAND_SOUND
  if (f < EXPAND_LIMIT)
    f -= (EXPAND_LIMIT-f)*(EXPAND_RATIO-1.0);
  if (f < 0.0)
    f = 0.0;
#endif
  return f;
}


/*

  SortSinusoids() changes the ordering of sine frequencies so as to minimize
  audible artifacts caused by changing sine frequencies and amplitudes. The
  idea is that the previous PlusV block is examined for each new sine value.
  If a frequency can be kept the same, sines are ordered so that this happens.
  If the change in the sine frequency is so big that it may be assumed that
  the sine is not from the same original frequency, but from a totally
  different harmonic, the old voice is first attenuated to zero amplitude
  before using the new value.

  Inputs:
    struct PlusV *v - PlusV structure with all fields set to valid values.

  Output:
    none

 */
static void SortSinusoids(struct PlusV *v) {
  /* These are all permutations of numbers 0, 1, 2, 3; no particular order */
  static const int perms[24][4] = {
    {0, 1, 2, 3},    {0, 1, 3, 2},
    {0, 2, 1, 3},    {0, 2, 3, 1},
    {0, 3, 1, 2},    {0, 3, 2, 1},
    {1, 0, 2, 3},    {1, 0, 3, 2},

    {1, 2, 0, 3},    {1, 3, 0, 2},
    {1, 2, 3, 0},    {1, 3, 2, 0},
    {2, 1, 0, 3},    {2, 1, 3, 0},
    {2, 0, 1, 3},    {2, 0, 3, 1},

    {2, 3, 1, 0},    {2, 3, 0, 1},
    {3, 0, 2, 1},    {3, 0, 1, 2},
    {3, 2, 0, 1},    {3, 1, 0, 2},
    {3, 2, 1, 0},    {3, 1, 2, 0}
  };
  static const int swaps[24][3][2] = {
    {{0, 0}, {0, 0}, {0, 0}},    {{2, 3}, {0, 0}, {0, 0}},
    {{1, 2}, {0, 0}, {0, 0}},    {{1, 2}, {2, 3}, {0, 0}},
    {{1, 2}, {1, 3}, {0, 0}},    {{1, 3}, {0, 0}, {0, 0}},
    {{0, 1}, {0, 0}, {0, 0}},    {{0, 1}, {2, 3}, {0, 0}},

    {{0, 1}, {1, 2}, {0, 0}},    {{0, 1}, {3, 2}, {1, 2}},
    {{0, 1}, {1, 2}, {2, 3}},    {{0, 1}, {1, 3}, {0, 0}},
    {{0, 2}, {0, 0}, {0, 0}},    {{0, 2}, {2, 3}, {0, 0}},
    {{0, 2}, {1, 2}, {0, 0}},    {{0, 2}, {1, 2}, {2, 3}},

    {{0, 2}, {1, 3}, {2, 3}},    {{0, 2}, {1, 3}, {0, 0}},
    {{0, 3}, {1, 3}, {0, 0}},    {{0, 3}, {1, 3}, {2, 3}},
    {{0, 3}, {1, 3}, {1, 2}},    {{0, 3}, {2, 3}, {0, 0}},
    {{0, 3}, {1, 2}, {0, 0}},    {{0, 3}, {0, 0}, {0, 0}}
  };
  flotari sums[24], minSum = 1e20;
  int i, j, minSumPos = -1, perfectMatches, maxPerfectMatches = 0;


  for (i=0; i<24; i++) {
    int ch;
    sums[i] = 0.0;
    perfectMatches = 0;
    for (j=0; j<4; j++) {
      flotari toAdd = 0.0, t;
      for (ch=0; ch<v->channels; ch++)
	toAdd += abs(v->oldSineEnergy[ch][j] -
		     v->sineEnergy[ch][perms[i][j]]);
      if (v->channels > 1)
	toAdd *= 0.5;
      toAdd += 1.0;
      t = abs(v->oldSineFreq[j] - v->sineFreq[perms[i][j]]);
      if (v->sines[perms[i][j]] && t < 3)
	perfectMatches++;
      toAdd *= t*t;


      sums[i] += toAdd;
    }
    if ((sums[i] < minSum && perfectMatches >= maxPerfectMatches) ||
	perfectMatches > maxPerfectMatches) {
      minSum = sums[i];
      minSumPos = i;
      maxPerfectMatches = perfectMatches;
    }
  }


  i = minSumPos;

  for (j=0; j<3; j++) {
    SwapSines(v, swaps[i][j][0], swaps[i][j][1]);
  }


  /* Now the sinusoids are sorted. It's our time to see if there are
     any stupidly big frequency jumps left. If there are, see if the
     voices can be shuffled around to avoid that. */

  for (i=0; i<4; i++) {
    if (v->oldSines[i] && v->sines[i] &&
	abs(v->oldSineFreq[i] - v->sineFreq[i]) > 5) {
      int j;
      for (j=0; j<4; j++) {
	/* If free sine, swap to it */
	if (!v->sines[j] && !v->oldSines[j]) {
	  SwapSines(v, i, j);
	  j = 10000;
	}
      }
      /* Regardless of if the change succeeded, clear the old sine */
      v->sines[i] = 0;
    }
  }
}



/*

  EncodeFile() converts a WAV file to two separate files: a WAV' file with
  fs' = fs/2, and a PlusV file. When this function has exited, the output
  WAV' file is to be forwarded to a MP3 (or other) encoder. Then the MP3
  file should be interleaved with the PlusV data output from this function.

  Inputs:
    const char *inName - Input WAV file name.
    const char *outName - Output WAV file name.
    const char *pvName - Output PlusV file name. If NULL, stdout is used.
    enum SND_FILETYPE outFileType - The filetype of the output audio file.
	F_WAVW is 16-bit WAV.
    int resample32 - Flag to tell if 44.1/48 kHz data is to be automatically
	resampled to 32 kHz before further processing.
    int lowBand - If negative, use typical 2:1 resampling. If not, don't
	resample, and lowBand is the lowest band handed over to PlusV.
	lowBand may in this case be between 0..7.

  Output:
    int - the destination WAV sample rate if successful, negative on failure.

 */
int EncodeFile(const char *inName, const char *outName, const char *pvName,
	       enum SND_FILETYPE outFileType, int resample32, int lowBand,
		   int ENCODE_BANDS, int ENCODE_QUANT, int TOTAL_ENCODER_DELAY) {
  s_int32 inRate = 0, outRate = 0;
  int i;
  int ch, chan;
  
  // [gameplaya15143] changed the array allocation to be dynamic :)
  //flotari eED[2][TOTAL_ENCODER_DELAY*2] = {{0}};	// <- original code
  // now my code..
  	// dynamically allocate the memory for this array (is used for the delay)
	flotari **eED;
	eED = calloc(2, sizeof(flotari));
	if(eED == NULL)
	{
		fprintf(stderr, "Couldn't allocate memory for eED array.\n");
		return 1;
	}
	for(i = 0; i < 2; i++)
	{
		eED[i] = calloc(TOTAL_ENCODER_DELAY * 2, sizeof(flotari));
		if(eED[i] == NULL)
		{
			fprintf(stderr, "Couldn't allocate memory for eED array: %d\n", i);
			return 1;
		}
	}	// end [gameplaya15143] mods
	
  int eEDPointer = 0;
  
  // [gameplaya15143] this one is now dynamic too :)
  //flotari highPassEnergyDelay[TOTAL_ENCODER_DELAY*2] = {0};
  	// and this one will be dynamically allocated as well ^^
	flotari *highPassEnergyDelay;
	highPassEnergyDelay = calloc(TOTAL_ENCODER_DELAY * 2, sizeof(flotari));
	if(highPassEnergyDelay == NULL)
	{
		fprintf(stderr, "Couldn't allocate memory for highPassEnergyDelay array.\n");
		return 1;
	}	// end [gameplaya15143] mods
  
  int hPEDPointer = 0;
  s_int16 rawSamples[2];
  flotari delaySamples[2], newSamples[2];
  FILE *readFP = NULL, *writeFP = NULL, *plusvFP = NULL;
  struct SAMPLE_HEADER *readSH = NULL, *writeSH = NULL;
  int success = 0;
  static flotari fftIn[2][FFT_BUF_SIZE], fftOut[2][FFT_BUF_SIZE];
  static flotari blackman[FFT_BUF_SIZE];
  int blockPos = 0;
  flotari energies[2][FFT_BUF_SIZE/2];
  int sines, encodeBands = 0;
  flotari sinEnergies[2][4];
  int sinFreqs[4], encodeResamp = 0, downsample = (lowBand < 0);
  long blockCount = 0;
  flotari crossOver[2][16] = {
    {
      0.04, 0.05, 0.07, 0.1,
      0.2, 0.25, 0.3, 0.35,
      0.4, 0.5, 0.6, 0.7,
      0.8, 0.9, 1.0, 1.0
    }, {
      1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 1, 1, 1, 1, 1, 1
    }
  };
  int trPos = 0, trNum = 0;
  flotari trEnergy[5] = {1.0, 1.0, 1.0, 1.0, 1.0};
  flotari totEnergy;
  int plusVBlockSize = 0;
  int fatBlockCount = 0;

  plusV.private =
#ifdef ALWAYS_ENCODE_SHORT_BLOCKS
    1
#else
    0
#endif
    ;

  fprintf(stderr, "Encode\n");
  BlackmanWindow(blackman, FFT_BUF_SIZE);

#if 0
  {
    unsigned long long loop = 0, looplim = 10000;
    extern unsigned long random_state;
    unsigned long startState = random_state;

    while (1) {
      if (random_state == startState) {
	fprintf(stderr, "WENT AROUND IN %lld LOOPS\n", loop);
      }
      if (loop++ >= looplim) {
	looplim = loop * 1.1;
	fprintf(stderr, "Ran %lld loops\r", loop); fflush(stderr);
      }
      pvrandom();
    }
  }
#endif

  ClearDecimationFilter();
  ClearLpFilter(lowBand);

  readFP = fopen(inName, "rb");
  if (!readFP) {
    fprintf(stderr, "Couldn't open \"%s\" for reading\n", inName);
    goto close_files;
  }

  readSH = ReadSampleHeader(readFP);
  if (!readSH) {
    fprintf(stderr, "Couldn't read sample header\n");
    goto close_files;
  }

  inRate = readSH->sampleRate;
  chan = readSH->channels;
  outRate = resample32 ? 16000 : (inRate / 2);

  switch(outRate) {
  case 16000:
#ifdef ENCODE_SHORT_BLOCKS_AT_32K
    plusV.private = 1;
#endif
    if (inRate == 44100)
      encodeResamp = 1;
    if (inRate == 48000)
      encodeResamp = 2;
//    encodeBands = ENCODE_BANDS_32;
    encodeBands = ENCODE_BANDS;	// [gameplaya15143]
    break;
  case 22050:
//    encodeBands = ENCODE_BANDS_32;
    encodeBands = ENCODE_BANDS;	// [gameplaya15143]
    break;
  case 24000:
//    encodeBands = ENCODE_BANDS_32;
    encodeBands = ENCODE_BANDS;	// [gameplaya15143]
    break;
  default:
    fprintf(stderr, "Sample rate %d is not in (32000, 44100, 48000)\n",
	    inRate);
    encodeBands = ENCODE_BANDS;	// [gameplaya15143] commented the next line too
//    goto close_files;	// now we can encode from sources with less than 32khz
  }

  if (chan > 2) {
    fprintf(stderr, "Must not have more than 2 channels ");
    goto close_files;
  }

  writeFP = fopen(outName, "wb");
  if (!writeFP) {
    fprintf(stderr, "Couldn't open \"%s\" for writing\n",
	    outName);
    goto close_files;
  }

  if (pvName) {
    plusvFP = fopen(pvName, "wb" );
    if(!plusvFP) {
      fprintf( stderr, "Couldn't open \"%s\" for writing\n", pvName );
      goto close_files;
    }
  } else {
    plusvFP = stdout;
  }

  if (!downsample)
    outRate *= 2;

  writeSH = WriteSampleHeader(writeFP, outFileType, chan, outRate, -1);

  success = 1;

  fprintf(stderr, " Input file is %d %s bits @ " FMT_S32 " Hz\n",
	  abs(readSH->bitsPerSample),
	  (readSH->bitsPerSample >= 0) ? "linear" : " u-law",
	  readSH->sampleRate);
  fprintf(stderr, "Output file is %d %s bits @ " FMT_S32 " Hz\n",
	  abs(writeSH->bitsPerSample),
	  (writeSH->bitsPerSample >= 0) ? "linear" : " u-law",
	  writeSH->sampleRate);
  fprintf(stderr, "" FMT_S32 " audio channel%s, "
	  "PlusV bands %d..%d ~= bandwidth %d..%d Hz\n",
	  chan, (chan != 1) ? "s" : "", (downsample?0:lowBand), encodeBands-1,
	  (8+(downsample?0:lowBand))*writeSH->sampleRate/(16*(downsample?1:2)),
	  (8+encodeBands)*writeSH->sampleRate/(16*(downsample?1:2)));

  plusVBlockSize = plusV.private ?
    PLUSV_SHORT_BLOCK_SIZE : PLUSV_LONG_BLOCK_SIZE;

  plusV.quant = ENCODE_QUANT;
  plusV.bands = encodeBands;
  plusV.version = ENCODE_VERSION;
  plusV.channels = chan;



  while (1) {
    switch (encodeResamp) {
    case 0:
      /* If no resampling is to be done, just read the two samples */
      if (ReadSamples(rawSamples, chan) < chan)
	goto close_files;
      for (ch=0; ch<chan; ch++) {
	/* The following line contains an implicit int-to-float conversion */
	newSamples[ch] = rawSamples[ch];
      }
      break;
    case 1:
      /* Read and convert 44.1 kHz to 32 kHz */
      do {
	if (ReadSamples(rawSamples, chan) < chan)
	  goto close_files;
	for (ch=0; ch<chan; ch++) {
	  /* The following line contains an implicit int-to-float conv. */
	  newSamples[ch] = rawSamples[ch];
	}
      } while (!Resample44to32(chan, newSamples));
      break;
    case 2:
      /* Read and convert 48 kHz to 32 kHz */
      do {
	if (ReadSamples(rawSamples, chan) < chan)
	  goto close_files;
	for (ch=0; ch<chan; ch++) {
	  /* The following line contains an implicit int-to-float conv. */
	  newSamples[ch] = rawSamples[ch];
	}
      } while (!Resample48to32(chan, newSamples));
      break;
    }
    for (ch=0; ch<chan; ch++) {
      delaySamples[ch] = eED[ch][eEDPointer];
      eED[ch][eEDPointer] = newSamples[ch];
    }

    if (blockPos >= plusVBlockSize - FFT_BUF_SIZE)
      for (ch=0; ch<chan; ch++)
	fftIn[ch][blockPos-(plusVBlockSize - FFT_BUF_SIZE)] =
	  delaySamples[ch];

    /*

      Decimation by 2

    */
    if (downsample) {
      static int sync = 0;
      if (++sync & 1) {
	/* For every other sample, just store values */
	DecimationFilterStore(chan, newSamples);
      } else {
	/* For every other sample, calculate new downsampled value */
	trEnergy[trNum+1] += highPassEnergyDelay[hPEDPointer];
	highPassEnergyDelay[hPEDPointer] = DecimationFilter(chan, newSamples);
	for (ch=0; ch<chan; ch++) {
	  if (fabs(newSamples[ch]) >= 32767.0) {
	    if (newSamples[ch] >= 0.0)
	      rawSamples[ch] = 32767;
	    else
	      rawSamples[ch] = -32768;
	  } else {
	    rawSamples[ch] = (s_int32)(newSamples[ch] + 40000.5) - 40000L;
	  }
	}
	/* Write low-pass filtered samples to WAV file */
#if 0
	rawSamples[0] = pvrandom();
	rawSamples[1] = 0;
#endif
	WriteSamples(rawSamples, chan);
	if (++hPEDPointer >= TOTAL_ENCODER_DELAY)
	  hPEDPointer = 0;
      }
    } else {
#if 1
      trEnergy[trNum+1] += highPassEnergyDelay[hPEDPointer];
#else
      trEnergy[trNum+1] += 1;
#endif
      highPassEnergyDelay[hPEDPointer] = LpFilter(chan, newSamples);
      for (ch=0; ch<chan; ch++) {
	if (fabs(newSamples[ch]) >= 32767.0) {
	  if (newSamples[ch] >= 0.0)
	    rawSamples[ch] = 32767;
	  else
	    rawSamples[ch] = -32768;
	} else {
	  rawSamples[ch] = (s_int32)(newSamples[ch] + 40000.5) - 40000L;
	}
      }
      /* Write low-pass filtered samples to WAV file */
      WriteSamples(rawSamples, chan);
      if (++hPEDPointer >= TOTAL_ENCODER_DELAY)
	hPEDPointer = 0;
    }
    
    if(++blockPos >= plusVBlockSize) {
      blockPos = 0;
      trPos = 0;
      trNum = 0;
    } else if (++trPos >= plusVBlockSize >> 2) {
      trPos = 0;
      trNum++;
    }
    if (++eEDPointer >=
	(downsample ? TOTAL_ENCODER_DELAY*2 : TOTAL_ENCODER_DELAY))
      eEDPointer = 0;

    /* If we have gone through a whole block, store results */
    if (!blockPos) {
      static int block = -1;
      flotari trAvg = 0, trMaxEnergyChange = -1.0, trMinEnergyChange = 1e8;
      int trMinLoc = -1;
      int trMaxLoc = -1;
      block++;
      /* Check if we want to make a transient or normal block */
      for (i=1; i<5; i++) {
	flotari t;
	t = trEnergy[i] / trEnergy[i-1];
	if (t > trMaxEnergyChange) {
	  trMaxEnergyChange = t;
	  trMaxLoc = i-1;
	}
	if (t < trMinEnergyChange) {
	  trMinEnergyChange = t;
	  trMinLoc = i-1;
	}
	trAvg += trEnergy[i];
      }

      /* If energy change is high enough and level is not too low, use
	 transient block */
      if (trMaxEnergyChange > TRANSIENT_ENERGY_UP_THRESHOLD &&
	  trMinEnergyChange > 2.0/TRANSIENT_ENERGY_DOWN_THRESHOLD &&
	  trEnergy[trMaxLoc+1] > 10000.0) {
#ifdef ENCODE_TRANSIENTS
	plusV.trans = 1;
#else
	plusV.trans = 0;
#endif
	plusV.trPos = trMaxLoc;
      } else if (trMinEnergyChange < 1.0/TRANSIENT_ENERGY_DOWN_THRESHOLD &&
		 trMaxEnergyChange < 0.5*TRANSIENT_ENERGY_UP_THRESHOLD &&
		 trEnergy[trMinLoc] > 10000.0) {
#ifdef ENCODE_TRANSIENTS
	plusV.trans = 3;
#else
	plusV.trans = 0;
#endif
	plusV.trPos = trMinLoc;
      } else {
	plusV.trans = 0;
      }

      totEnergy = 0.0;
      for (ch=0; ch<chan; ch++) {
	for (i=0; i<FFT_BUF_SIZE; i++) {
	  fftOut[ch][i] = fftIn[ch][i] * blackman[i] * (1.0/8192.0);
	}

	fft512(fftOut[ch]);

#if 0
	for (i=0; i<256; i++) {
	  printf("%3d %14.5f %14.5f\n", i, fftOut[ch][i], tmpfft[i*2]);
	  printf("%3d %14.5f %14.5f\n", i, fftOut[ch][512-i], tmpfft[i*2+1]);
	}
#endif

	/* If the FFT block is longer than our block length, shift a bit. */
	/* Note: For MP3 this is true if short blocks are used. */
	if (plusVBlockSize < FFT_BUF_SIZE)
	  for (i=0; i<FFT_BUF_SIZE-plusVBlockSize; i++)
	    fftIn[ch][i] = fftIn[ch][i+plusVBlockSize];

	/* Collect energies */
	for (i=0; i<FFT_BUF_SIZE/4-8; i++)
	  energies[ch][i] = 0;
	for (i=FFT_BUF_SIZE/4; i<FFT_BUF_SIZE/2; i++) {
	  flotari R = fftOut[ch][i*2  ];
	  flotari I = fftOut[ch][i*2+1];
	  totEnergy += (energies[ch][i] = R*R+I*I);
	}
      }


      /* Sharpen transient edges by correcting energy levels */
      if (
#ifdef SHARPEN_TRANSIENTS
	  plusV.trans
#else
	  0
#endif
	  ) {
	flotari ratio;
	totEnergy *= 150*576/4;	/* '150' is an empirically tested value */
	if (plusV.trans == 1) {
	  ratio = trEnergy[trMaxLoc+1]/totEnergy;
	} else {
	  ratio = trEnergy[trMinLoc+1]/totEnergy;
	}
	if ((ratio > 1.0 && plusV.trans == 1) ||
	    (ratio < 1.0 && plusV.trans > 1)) {

	  for (ch=0; ch<chan; ch++) {
	    for (i=0; i<FFT_BUF_SIZE/2; i++) {
	      energies[ch][i] *= ratio;
	    }
	  }
	}
      }


      trEnergy[0] = trEnergy[4];
      for (i=1; i<5; i++)
	trEnergy[i] = 1.0;

      for (ch=0; ch<chan; ch++) {
	for (i=0; i<16; i++)
	  energies[ch][FFT_BUF_SIZE/4-8+i] *= crossOver[downsample][i];
      }



      {
	flotari chLocalEnergies[2][FFT_BUF_SIZE/2];
	flotari localEnergies[FFT_BUF_SIZE/2];
	int usedSinFreq[FFT_BUF_SIZE/2];
	static int lastUsedSinFreq[FFT_BUF_SIZE/2] = {0};

	sines = 0;
#ifdef ENCODE_SINES
	/* Find and remove sinusoidal signals */
	/*
	  Note, that this algorithm is incredible simplistic and
	  outright stupid. In some cases it may cause audio quality to
	  be lessened considerably although in many cases it enhances
	  quality. With a better sinusoid seeker, audio quality would
	  definitely be enhanced.
	*/
	for (i=0; i<FFT_BUF_SIZE/2; i++) {
	  usedSinFreq[i] = 0;
	  localEnergies[i] = 0;
	  for (ch=0; ch<chan; ch++) {
	    localEnergies[i] += energies[ch][i];
	    chLocalEnergies[ch][i] = energies[ch][i];
	  }
	}

	for (i=0; i<4; i++) {
	  flotari maxEnergy = -1.0, average;
	  int maxF = -1, j;

	  for (j=FFT_BUF_SIZE/4+FFT_BUF_SIZE*lowBand/32;
	       j<FFT_BUF_SIZE/4+FFT_BUF_SIZE*encodeBands/32; j++) {
	    flotari lE;
	    average =
	      (localEnergies[j-3] + localEnergies[j+3] +
	       localEnergies[j-4] + localEnergies[j+4] +
	       localEnergies[j-5] + localEnergies[j+5])*(1.0/6.0);
	    lE = localEnergies[j];
#if 1
	    if (lastUsedSinFreq[j])
	      lE *= 2.0;
#endif
	    if (lE > maxEnergy && lE > average*SIN_THRESHOLD &&
		!usedSinFreq[j]) {
	      maxEnergy = lE;
	      maxF = j;
	    }
	  }


	  if (maxF >= FFT_BUF_SIZE/4) {
	    sines = i+1;
	    sinFreqs[i] = maxF-FFT_BUF_SIZE/4;
	    for (ch=0; ch<chan; ch++) {
	      sinEnergies[ch][i] = 0;
	      /* Clear sin energy from original energies to prevent
		 same information to be coded as noise. */
	      for (j=maxF-2; j<=maxF+2; j++) {
		sinEnergies[ch][i] += chLocalEnergies[ch][j];
		energies[ch][j] = 0.0;
		/* Yes, the following table may be cleared twice. */
		usedSinFreq[j] = 1;
	      }
	    }
	  } else {
	    /* End loop */
	    i = 10000;
	  }
	}

	/* Mark which sines were used for further calculation */
	for (i=FFT_BUF_SIZE/4; i<FFT_BUF_SIZE/2; i++) {
	  lastUsedSinFreq[i] = (usedSinFreq[i] ? 1 : 0);
	}

#endif


	/* Store sinusoids */
#if 0
	sines = 0;
#endif
	for (ch=0; ch<chan; ch++) {
	  for (i=0; i<sines; i++) {
	    if (sinEnergies[ch][i] > 0.0001) {
	      sinEnergies[ch][i] = (20.0*log10(sqrt(sinEnergies[ch][i])*
					       (1.414 / FFT_BUF_SIZE))+94) *
		(1.0/DB_CORR);
	    }
	    if (sinEnergies[ch][i] < 0.2) {
	      sinEnergies[ch][i] = 0.0;
	    }
	  }
	}

	for (i=0; i<sines; i++) {
	  flotari sE = -1.0;
	  for (ch=0; ch<chan; ch++)
	    sE = max(sE, sinEnergies[ch][i]);
	  if (sE < 0.2) {
	    sines = i;
	    i = 10000;
	  }
	}

	for (i=0; i<sines; i++) {
	  plusV.sines[i] = 1;
	  plusV.sineFreq[i] = sinFreqs[i];

	  for (ch=0; ch<chan; ch++) {
	    plusV.sineEnergy[ch][i] =
	      RoundEnergy(Expand(sinEnergies[ch][i]), ENCODE_QUANT);
	  }
	}

	for (i=sines; i<4; i++) {
	  plusV.sines[i] = 0;
	  for (ch=0; ch<chan; ch++)
	    plusV.sineEnergy[ch][i] = 0;
	}
      }

      /* Sort sinusoids. This is important to get the sinusoids in
         a sensible order. Otherwise there may be weird sounding chirping
         here and there. */
      SortSinusoids(&plusV);

      for (ch=0; ch<chan; ch++) {
	/* Store energy bands */
	for (i=0; i<lowBand; i++)
	  plusV.noiseEnergy[ch][i] = 0.0;

	for (i=FFT_SLICES/2+lowBand; i<FFT_SLICES; i++) {
	  flotari energy = 0.0;
	  int j;
	  for (j=0; j<FFT_SLICE_SIZE; j++)
	    energy += energies[ch][i*FFT_SLICE_SIZE + j];

	  if (energy > 0.001)
	    energy = (20.0 * log10(sqrt(energy)*(1.4142 / FFT_BUF_SIZE)) + 94)*
	      (1.0/DB_CORR);
	  if (energy < 0.0)
	    energy = 0.0;

	  plusV.noiseEnergy[ch][i-FFT_SLICES/2] =
	    RoundEnergy(Expand(energy), ENCODE_QUANT);
	}
      }

      /* Finally, write the calculated block */
      WritePlusVBlock(plusvFP, &plusV,
#ifdef ENCODE_ONLY_FAT_BLOCKS
		      1
#else
		      !fatBlockCount
#endif
		      );
      if (++fatBlockCount >= FAT_BLOCK_LIMIT)
	fatBlockCount = 0;
      blockCount++;
    }
  } /* while (1) */

  /* Cleanup */
close_files:
  if(plusvFP) {
#ifdef NO_MAIN
    extern int pvbitspersec;
#else
    int pvbitspersec;
#endif

    pvbitspersec = (int)(ftell(plusvFP)*8 /
			 (blockCount * plusVBlockSize /
			  (writeSH->sampleRate*(downsample ? 2.0 : 1.0)))+0.5);

    fprintf(stderr, "Wrote %2.1f seconds (%ld blocks) at %d bits/s\n",
	    blockCount * plusVBlockSize / (writeSH->sampleRate *
					   (downsample ? 2.0 : 1.0)),
	    blockCount, pvbitspersec );

    if (plusvFP != stdout)
      fclose(plusvFP);
  }

  if (writeFP) {
    WriteSampleFooter();
    fclose(writeFP);
  }

  if(readFP)
    fclose(readFP);

  return success ? outRate : -1;
}





struct AudioMemMono {
  s_int32 sinFreq[4], lastSinFreq[4];
  s_int32 currSinFreq[4], currSinFreqDelta[4];
  u_int16 sinPhase[4];
};

struct AudioMemStereo {
  u_int32 noiseAmp[8], currNoiseAmp[8];
  s_int32 currNoiseAmpDelta[8];
  s_int32 sinAmp[4];
  s_int32 currSinAmp[4], currSinAmpDelta[4];
};

struct AudioMem {
  struct AudioMemMono m;
  struct AudioMemStereo s[2];
};

struct AudioMem aM;

/*

  DecodeFile() reads a WAV audio file and a PlusV bitstream that has been
  extracted from an MP3 file. It upsamples the audio file and joins PlusV
  data, writing its output as a new WAV file.

  Inputs:
    const char *inName - Input WAV file name.
    const char *outName - Output WAV file name.
    const char *pvName - Input PlusV file name. If NULL, stdin is used.
    enum SND_FILETYPE outFileType - The filetype of the output audio file.
	F_WAVW is 16-bit WAV.
    int resample32 - Flag to tell if 44.1/48 kHz data is to be automatically
	resampled to 32 kHz before further processing.

  Output:
    int - EXIT_SUCCESS if successful, otherwise EXIT_FAILURE

 */
int DecodeFile(const char *inName, const char *outName, const char *pvName,
	       enum SND_FILETYPE outFileType) {
  s_int32 inRate = 0, outRate = -1;
  u_int32 i;
  s_int32 ch, chan;
  s_int16 tmpSamples[2];
  flotari samples[2];
  flotari ubSamples[2];
  FILE *readFP = NULL, *writeFP = NULL, *plusvFP = NULL;
  struct SAMPLE_HEADER *readSH = NULL, *writeSH = NULL;
  s_int32 success = 0;
  int blockPos = 0;
  int trPos = 0, trNum = 0;
  int transient[4];
  int plusVBlockSize = 0;
  int upsample = 0;
  struct FMem *fm;

  fprintf(stderr, "Decode\n");

  i = ClearInterpolationFilter();
  if (!(fm = InitDelay(2, i))) {
    fprintf(stderr, "Out of memory!\n");
    goto close_files;
  }
  ClearUpperHalfIntFilter();
  ClearEightBandFilter();
  for (ch=0; ch<2; ch++) {
    for (i=0; i<8; i++) {
      aM.s[ch].noiseAmp[i] = 0;
      aM.s[ch].currNoiseAmp[i] = 0;
      aM.s[ch].currNoiseAmpDelta[i] = 0;
    }
    for (i=0; i<4; i++) {
      aM.s[ch].currSinAmp[i] = 0;
      aM.s[ch].currSinAmpDelta[i] = 0;
      aM.s[ch].sinAmp[i] = 0;
    }
  }
  for (i=0; i<4; i++) {
    aM.m.currSinFreq[i] = 0;
    aM.m.currSinFreqDelta[i] = 0;
    aM.m.sinFreq[i] = 0;
    aM.m.lastSinFreq[i] = 0;
    aM.m.sinPhase[i] = 0;
  }

  readFP = fopen(inName, "rb");
  if (!readFP) {
    fprintf(stderr, "Couldn't open \"%s\" for reading\n", inName);
    goto close_files;
  }

  readSH = ReadSampleHeader(readFP);
  if (!readSH) {
    fprintf(stderr, "Couldn't read sample header\n");
    goto close_files;
  }
  inRate = readSH->sampleRate;
  chan = readSH->channels;
  if (inRate > 24000) {
    upsample = 0;
    outRate = inRate;
  } else {
    upsample = 1;
    outRate = inRate * 2;
  }

  if (chan > 2) {
    fprintf(stderr, "must not have more than 2 channels ");
    goto close_files;
  }

  if (pvName) {
    plusvFP = fopen(pvName, "rb" );
    if(!plusvFP) {
      fprintf( stderr, "Couldn't open \"%s\" for reading\n", pvName );
      goto close_files;
    }
  } else {
    plusvFP = stdin;
  }

  writeFP = fopen(outName, "wb");
  if (!writeFP) {
    fprintf(stderr, "Couldn't open \"%s\" for writing\n", outName);
    goto close_files;
  }

  writeSH = WriteSampleHeader(writeFP, outFileType, chan, outRate, -1);

  success = 1;

  fprintf(stderr, " Input file is %d %s bits @ " FMT_S32 " Hz\n",
	  abs(readSH->bitsPerSample),
	  (readSH->bitsPerSample >= 0) ? "linear" : " u-law",
	  readSH->sampleRate);
  fprintf(stderr, "Output file is %d %s bits @ " FMT_S32 " Hz\n",
	  abs(writeSH->bitsPerSample),
	  (writeSH->bitsPerSample >= 0) ? "linear" : " u-law",
	  writeSH->sampleRate);
  fprintf(stderr, "" FMT_S32 " audio channel%s\n",
	  chan, (chan != 1) ? "s" : "");

  /* Read one mono or stereo sample at the time */
  while (ReadSamples(tmpSamples, chan) == chan) {
    for (ch=0; ch<chan; ch++)
      samples[ch] = tmpSamples[ch];

    /* If it is time to read a new block, do so */
    if (!blockPos) {
      for (ch=0; ch<chan; ch++)
	for (i=0; i<4; i++)
	  aM.s[ch].currSinAmp[i] = aM.s[ch].sinAmp[i] << 16;

      if (ReadPlusVBlock(plusvFP, &plusV) || plusV.channels != chan) {
	fprintf(stderr, "Cannot read auxilliary data\n");
	for (ch=0; ch<2; ch++) {
	  for (i=0; i<4; i++) {
	    plusV.sines[i] = 0;
	    plusV.sineFreq[i] = 0;
	    plusV.sineEnergy[ch][i] = 0;
	  }
	  for (i=0; i<8; i++)
	    plusV.noiseEnergy[ch][i] = 0;
	}
	plusV.private = 0;
      }

      plusVBlockSize = plusV.private ?
	PLUSV_SHORT_BLOCK_SIZE : PLUSV_LONG_BLOCK_SIZE;
      trPos = 0;
      trNum = 0;
      for (i=0; i<4; i++)
	transient[i] = plusV.trans ? (i == plusV.trPos) : 1;

      /* Filter bank -based way of calculating noise & sines at 22 kHz */
      for (i=0; i<4; i++) {
	aM.m.currSinFreq[i] = aM.m.sinFreq[i]*65536;
	if (plusV.sines[i]) {
	  aM.m.sinFreq[i] = (65536/2/128)*(128-plusV.sineFreq[i]);
	  /* If old amplitude == 0, jump straight to new frequency */
	  if (!aM.s[0].currSinAmp[i])
	    aM.m.currSinFreq[i] = aM.m.sinFreq[i] << 16;
	  if (aM.m.sinFreq[i] == 32768U)
	    aM.m.sinPhase[i] = 16384;
	}
	aM.m.currSinFreqDelta[i] =
	  (aM.m.sinFreq[i]*65536-aM.m.currSinFreq[i])/
	  (plusVBlockSize >> 1)*
	  ((plusV.trans) ? 4 : 1);
	if (aM.m.currSinFreqDelta[i] < 0)
	  aM.m.currSinFreqDelta[i]++;
      }

      for (ch=0; ch<chan; ch++) {
	for (i=0; i<8; i++) {
	  aM.s[ch].currNoiseAmp[i] = aM.s[ch].noiseAmp[i] * 65536;
	  if (plusV.noiseEnergy[ch][i]) {
	    flotari t =
	      pow(10.0,(plusV.noiseEnergy[ch][i]-188)/40.0 * DB_CORR);
	    if (t > 0.999)
	      t = 0.999;
	    aM.s[ch].noiseAmp[i] = t * 65536;
	  } else {
	    aM.s[ch].noiseAmp[i] = 0;
	  }
	  aM.s[ch].currNoiseAmpDelta[i] =
	    ((s_int32)(aM.s[ch].noiseAmp[i]*32768)
	     -(s_int32)(aM.s[ch].currNoiseAmp[i]/2)) /
	    (plusVBlockSize >> 2) * ((plusV.trans) ? 4 : 1);
	  if (aM.s[ch].currNoiseAmpDelta[i] < 0)
	    aM.s[ch].currNoiseAmpDelta[i]++;
#if 0
	  printf("NA[%d] = %6d, cNA = %10d >> 16 = %6d, delta=%10d\n", i,
		 aM[ch].noiseAmp[i],
		 aM[ch].currNoiseAmp[i], aM[ch].currNoiseAmp[i]>>16,
		 aM[ch].currNoiseAmpDelta[i]);
#endif
	}
	for (i=0; i<4; i++) {
	  aM.s[ch].currSinAmp[i] = aM.s[ch].sinAmp[i] << 16;
	  if (plusV.sines[i]) {
	    aM.s[ch].sinAmp[i] = pow(10.0, (plusV.sineEnergy[ch][i]-188)/40.0 *
				   DB_CORR)* 8192.0;
	  } else {
	    aM.s[ch].sinAmp[i] = 0;
	  }
	  aM.s[ch].currSinAmpDelta[i] =
	    (aM.s[ch].sinAmp[i]*65536-aM.s[ch].currSinAmp[i])/
	    (plusVBlockSize >> 1)*
	    ((plusV.trans) ? 4 : 1);
	  if (aM.s[ch].currSinAmpDelta[i] < 0)
	    aM.s[ch].currSinAmpDelta[i]++;
	}
      }
    } /* if (!blockPos) */


    {
      flotari ebSamples[8];

      ebSamples[0] = ((pvrandom()&0xFFFF)-32768) * 63500 / 65536.0;

      EightBandFilter(ebSamples);

#ifdef DECODE_SINS
	for (i=0; i<4; i++) {
	  aM.m.sinPhase[i] += aM.m.currSinFreq[i]>>16;
	  if (transient[trNum])
	    aM.m.currSinFreq[i] += aM.m.currSinFreqDelta[i];
	}
#endif

      for (ch=0; ch<chan; ch++) {
	ubSamples[ch] = 0.0;
#ifdef DECODE_NOISES
	for (i=0; i<8; i++) {
	  ubSamples[ch] += ebSamples[7-i] * (aM.s[ch].currNoiseAmp[i]>>16) /
	    65536;
#if 0
	  if (i==2)
	    printf("cNA[%d] = %10d, d=%8d\n", i,
		   aM[ch].currNoiseAmp[i], aM[ch].currNoiseAmpDelta[i]);
#endif
	  if (transient[trNum])
	    aM.s[ch].currNoiseAmp[i] += aM.s[ch].currNoiseAmpDelta[i];
	}
#endif
#ifdef DECODE_SINS
	for (i=0; i<4; i++) {
	  ubSamples[ch] += (aM.s[ch].currSinAmp[i]>>16) *
	    sin(aM.m.sinPhase[i] * (2.0*M_PI/65536.0));
	  if (transient[trNum]) {
	    aM.s[ch].currSinAmp[i] += aM.s[ch].currSinAmpDelta[i];
	  }
	}
#endif
      }
    }

    /*
      Now we have two signals. Let's say fs = sample rate of final signal.
      "samples" contains the lower half of the spectrum, sampled at fs/2.
      "ubSamples" contains the higher half of the spectrum, also sampled at
      fs/2. Now, we upsample both signal. We low-pass filter "samples" and
      high-pass filter "ubSamples", thus getting two clean, non-overlapping
      audio portions that may simply be added to each other.
    */

    /* Upsample by 2 */
    for (i=0; i<2; i++) {
      flotari tmpVal;

      if (upsample) {
	InterpolationFilter(chan, samples, i);
      } else {
	if (i == 1) {
	  ReadSamples(tmpSamples, chan);
	  for (ch=0; ch<chan; ch++)
	    samples[ch] = tmpSamples[ch];
	}
	FilterDelay(fm, samples, chan);
      }
      UpperHalfIntFilter(chan, ubSamples, i);

      for (ch=0; ch<chan; ch++) {
	tmpVal =
#ifdef DECODE_LOWER
	  samples[ch] +
#endif
#ifdef DECODE_UPPER
	  ubSamples[ch];
#else
	  0;
#endif

#ifdef MARK_DECODE_BLOCK_LIMITS
      if (!blockPos && !ch)
	tmpVal = 2000;
#endif
#ifdef MARK_TRANSIENT_BLOCKS
      if (!blockPos && plusV.trans && !ch)
	tmpVal = 8000;
#endif

	if (fabs(tmpVal) >= 32767.0) {
	  if (tmpVal >= 32767.0)
	    tmpSamples[ch] = 32767;
	  else
	    tmpSamples[ch] = -32768;
	} else {
	  tmpSamples[ch] = (s_int32)(tmpVal + 40000.5) - 40000L;
	}
      }
      WriteSamples(tmpSamples, chan);
    }

    if ((blockPos += 2) >= plusVBlockSize)
      blockPos = 0;
    if ((trPos += 2) >= plusVBlockSize/4) {
      trPos = 0;
      trNum++;
    }
  } /* while (could read samples) */

close_files:
  if (fm) {
    FreeFilterDelay(fm);
  }

  if (writeFP) {
    WriteSampleFooter();
    fclose(writeFP);
  }

  if (readFP)
    fclose(readFP);

  if (plusvFP)
    fclose(plusvFP);

  return success ? EXIT_SUCCESS : EXIT_FAILURE;
}
